@@ -31,6 +31,8 @@
 #include <proto/stick_table.h>
 #include <proto/task.h>
 
+#define IS_HTX_STRM(strm) ((strm)->flags & SF_HTX)
+
 extern struct pool_head *pool_head_stream;
 extern struct list streams;
 
@@ -41,6 +43,8 @@ int stream_create_from_cs(struct conn_stream *cs);
 
 /* kill a stream and set the termination flags to <why> (one of SF_ERR_*) */
 void stream_shutdown(struct stream *stream, int why);
+void stream_dump(struct buffer *buf, const struct stream *s, const char *pfx, char eol);
+void stream_dump_and_crash(enum obj_type *obj, int rate);
 
 void stream_process_counters(struct stream *s);
 void sess_change_server(struct stream *sess, struct server *newsrv);
@@ -202,7 +206,7 @@ static inline void stream_track_stkctr(struct stkctr *ctr, struct stktable *t, s
 }
 
 /* Increase the number of cumulated HTTP requests in the tracked counters */
-static void inline stream_inc_http_req_ctr(struct stream *s)
+static inline void stream_inc_http_req_ctr(struct stream *s)
 {
 	struct stksess *ts;
 	void *ptr;
@@ -240,7 +244,7 @@ static void inline stream_inc_http_req_ctr(struct stream *s)
 /* Increase the number of cumulated HTTP requests in the backend's tracked
  * counters. We don't look up the session since it cannot happen in the bakcend.
  */
-static void inline stream_inc_be_http_req_ctr(struct stream *s)
+static inline void stream_inc_be_http_req_ctr(struct stream *s)
 {
 	struct stksess *ts;
 	void *ptr;
@@ -280,7 +284,7 @@ static void inline stream_inc_be_http_req_ctr(struct stream *s)
  * Note that even 404 are interesting because they're generally caused by
  * vulnerability scans.
  */
-static void inline stream_inc_http_err_ctr(struct stream *s)
+static inline void stream_inc_http_err_ctr(struct stream *s)
 {
 	struct stksess *ts;
 	void *ptr;
@@ -315,20 +319,20 @@ static void inline stream_inc_http_err_ctr(struct stream *s)
 	}
 }
 
-static void inline __stream_add_srv_conn(struct stream *sess, struct server *srv)
+static inline void __stream_add_srv_conn(struct stream *sess, struct server *srv)
 {
 	sess->srv_conn = srv;
 	LIST_ADD(&srv->actconns, &sess->by_srv);
 }
 
-static void inline stream_add_srv_conn(struct stream *sess, struct server *srv)
+static inline void stream_add_srv_conn(struct stream *sess, struct server *srv)
 {
 	HA_SPIN_LOCK(SERVER_LOCK, &srv->lock);
 	__stream_add_srv_conn(sess, srv);
 	HA_SPIN_UNLOCK(SERVER_LOCK, &srv->lock);
 }
 
-static void inline stream_del_srv_conn(struct stream *sess)
+static inline void stream_del_srv_conn(struct stream *sess)
 {
 	struct server *srv = sess->srv_conn;
 
@@ -341,13 +345,14 @@ static void inline stream_del_srv_conn(struct stream *sess)
 	HA_SPIN_UNLOCK(SERVER_LOCK, &srv->lock);
 }
 
-static void inline stream_init_srv_conn(struct stream *sess)
+static inline void stream_init_srv_conn(struct stream *sess)
 {
 	sess->srv_conn = NULL;
 	LIST_INIT(&sess->by_srv);
 }
 
 void service_keywords_register(struct action_kw_list *kw_list);
+void list_services(FILE *out);
 
 #endif /* _PROTO_STREAM_H */
 
