@@ -77,6 +77,7 @@ enum act_name {
 	ACT_HTTP_DEL_ACL,
 	ACT_HTTP_DEL_MAP,
 	ACT_HTTP_SET_MAP,
+	ACT_HTTP_EARLY_HINT,
 
 	/* http request actions. */
 	ACT_HTTP_REQ_TARPIT,
@@ -106,6 +107,13 @@ struct act_rule {
 	struct action_kw *kw;
 	struct applet applet;                  /* used for the applet registration. */
 	union {
+		struct {
+			struct sample_expr *expr;
+			char *varname;
+			char *resolvers_id;
+			struct dns_resolvers *resolvers;
+			struct dns_options dns_opts;
+		} dns;                        /* dns resolution */
 		struct {
 			char *realm;
 		} auth;                        /* arg used by "auth" */
@@ -113,8 +121,13 @@ struct act_rule {
 			char *name;            /* header name */
 			int name_len;          /* header name's length */
 			struct list fmt;       /* log-format compatible expression */
-			struct my_regex re;    /* used by replace-header and replace-value */
+			struct my_regex *re;   /* used by replace-header and replace-value */
 		} hdr_add;                     /* args used by "add-header" and "set-header" */
+		struct {
+			char *name;            /* header name */
+			int name_len;          /* header name's length */
+			struct list fmt;       /* log-format compatible expression */
+		} early_hint;
 		struct redirect_rule *redir;   /* redirect rule or "http-request redirect" */
 		int nice;                      /* nice value for ACT_HTTP_SET_NICE */
 		int loglevel;                  /* log-level value for ACT_HTTP_SET_LOGL */
