@@ -24,6 +24,7 @@
 #include <common/config.h>
 #include <common/debug.h>
 #include <common/hash.h>
+#include <common/htx.h>
 #include <common/initcall.h>
 #include <common/ticks.h>
 #include <common/time.h>
@@ -36,6 +37,7 @@
 #include <proto/backend.h>
 #include <proto/channel.h>
 #include <proto/frontend.h>
+#include <proto/http_htx.h>
 #include <proto/lb_chash.h>
 #include <proto/lb_fas.h>
 #include <proto/lb_fwlc.h>
@@ -55,11 +57,8 @@
 #include <proto/session.h>
 #include <proto/stream.h>
 #include <proto/stream_interface.h>
-#include <proto/task.h>
-
-#ifdef USE_OPENSSL
 #include <proto/ssl_sock.h>
-#endif /* USE_OPENSSL */
+#include <proto/task.h>
 
 int be_lastsession(const struct proxy *be)
 {
@@ -185,7 +184,7 @@ static struct server *get_server_sh(struct proxy *px, const char *addr, int len,
 	if ((px->lbprm.algo & BE_LB_HASH_MOD) == BE_LB_HMOD_AVAL)
 		h = full_hash(h);
  hash_done:
-	if (px->lbprm.algo & BE_LB_LKUP_CHTREE)
+	if ((px->lbprm.algo & BE_LB_LKUP) == BE_LB_LKUP_CHTREE)
 		return chash_get_server_hash(px, h, avoid);
 	else
 		return map_get_server_hash(px, h);
@@ -197,7 +196,9 @@ static struct server *get_server_sh(struct proxy *px, const char *addr, int len,
  * ends at the question mark. Depending on the number of active/backup servers,
  * it will either look for active servers, or for backup servers.
  * If any server is found, it will be returned. If no valid server is found,
- * NULL is returned.
+ * NULL is returned. The lbprm.arg_opt{1,2,3} values correspond respectively to
+ * the "whole" optional argument (boolean), the "len" argument (numeric) and
+ * the "depth" argument (numeric).
  *
  * This code was contributed by Guillaume Dallaire, who also selected this hash
  * algorithm out of a tens because it gave him the best results.
@@ -217,18 +218,18 @@ static struct server *get_server_uh(struct proxy *px, char *uri, int uri_len, co
 	if (px->lbprm.tot_used == 1)
 		goto hash_done;
 
-	if (px->uri_len_limit)
-		uri_len = MIN(uri_len, px->uri_len_limit);
+	if (px->lbprm.arg_opt2) // "len"
+		uri_len = MIN(uri_len, px->lbprm.arg_opt2);
 
 	start = end = uri;
 	while (uri_len--) {
 		c = *end;
 		if (c == '/') {
 			slashes++;
-			if (slashes == px->uri_dirs_depth1) /* depth+1 */
+			if (slashes == px->lbprm.arg_opt3) /* depth+1 */
 				break;
 		}
-		else if (c == '?' && !px->uri_whole)
+		else if (c == '?' && !px->lbprm.arg_opt1) // "whole"
 			break;
 		end++;
 	}
@@ -238,7 +239,7 @@ static struct server *get_server_uh(struct proxy *px, char *uri, int uri_len, co
 	if ((px->lbprm.algo & BE_LB_HASH_MOD) == BE_LB_HMOD_AVAL)
 		hash = full_hash(hash);
  hash_done:
-	if (px->lbprm.algo & BE_LB_LKUP_CHTREE)
+	if ((px->lbprm.algo & BE_LB_LKUP) == BE_LB_LKUP_CHTREE)
 		return chash_get_server_hash(px, hash, avoid);
 	else
 		return map_get_server_hash(px, hash);
@@ -271,13 +272,13 @@ static struct server *get_server_ph(struct proxy *px, const char *uri, int uri_l
 	p++;
 
 	uri_len -= (p - uri);
-	plen = px->url_param_len;
+	plen = px->lbprm.arg_len;
 	params = p;
 
 	while (uri_len > plen) {
 		/* Look for the parameter name followed by an equal symbol */
 		if (params[plen] == '=') {
-			if (memcmp(params, px->url_param_name, plen) == 0) {
+			if (memcmp(params, px->lbprm.arg_str, plen) == 0) {
 				/* OK, we have the parameter here at <params>, and
 				 * the value after the equal sign, at <p>
 				 * skip the equal symbol
@@ -295,7 +296,7 @@ static struct server *get_server_ph(struct proxy *px, const char *uri, int uri_l
 				if ((px->lbprm.algo & BE_LB_HASH_MOD) == BE_LB_HMOD_AVAL)
 					hash = full_hash(hash);
 
-				if (px->lbprm.algo & BE_LB_LKUP_CHTREE)
+				if ((px->lbprm.algo & BE_LB_LKUP) == BE_LB_LKUP_CHTREE)
 					return chash_get_server_hash(px, hash, avoid);
 				else
 					return map_get_server_hash(px, hash);
@@ -318,29 +319,51 @@ static struct server *get_server_ph(struct proxy *px, const char *uri, int uri_l
 static struct server *get_server_ph_post(struct stream *s, const struct server *avoid)
 {
 	unsigned int hash = 0;
-	struct http_txn *txn  = s->txn;
 	struct channel  *req  = &s->req;
-	struct http_msg *msg  = &txn->req;
 	struct proxy    *px   = s->be;
-	unsigned int     plen = px->url_param_len;
-	unsigned long    len  = http_body_bytes(msg);
-	const char      *params = c_ptr(req, -http_data_rewind(msg));
-	const char      *p    = params;
-	const char      *start, *end;
-
-	if (len == 0)
-		return NULL;
-
-	if (len > b_wrap(&req->buf) - p)
-		len = b_wrap(&req->buf) - p;
+	unsigned int     plen = px->lbprm.arg_len;
+	unsigned long    len;
+	const char      *params, *p, *start, *end;
 
 	if (px->lbprm.tot_weight == 0)
 		return NULL;
 
+	if (!IS_HTX_STRM(s)) {
+		struct http_txn *txn = s->txn;
+		struct http_msg *msg = &txn->req;
+
+		len  = http_body_bytes(msg);
+		p = params = c_ptr(req, -http_data_rewind(msg));
+
+		if (len == 0)
+			return NULL;
+		if (len > b_wrap(&req->buf) - p)
+			len = b_wrap(&req->buf) - p;
+
+	}
+	else {
+		struct htx *htx = htxbuf(&req->buf);
+		struct htx_blk *blk;
+
+		p = params = NULL;
+		len = 0;
+		for (blk = htx_get_first_blk(htx); blk; blk = htx_get_next_blk(htx, blk)) {
+			enum htx_blk_type type = htx_get_blk_type(blk);
+			struct ist v;
+
+			if (type != HTX_BLK_DATA)
+				continue;
+			v = htx_get_blk_value(htx, blk);
+			p = params = v.ptr;
+			len = v.len;
+			break;
+		}
+	}
+
 	while (len > plen) {
 		/* Look for the parameter name followed by an equal symbol */
 		if (params[plen] == '=') {
-			if (memcmp(params, px->url_param_name, plen) == 0) {
+			if (memcmp(params, px->lbprm.arg_str, plen) == 0) {
 				/* OK, we have the parameter here at <params>, and
 				 * the value after the equal sign, at <p>
 				 * skip the equal symbol
@@ -369,7 +392,7 @@ static struct server *get_server_ph_post(struct stream *s, const struct server *
 				if ((px->lbprm.algo & BE_LB_HASH_MOD) == BE_LB_HMOD_AVAL)
 					hash = full_hash(hash);
 
-				if (px->lbprm.algo & BE_LB_LKUP_CHTREE)
+				if ((px->lbprm.algo & BE_LB_LKUP) == BE_LB_LKUP_CHTREE)
 					return chash_get_server_hash(px, hash, avoid);
 				else
 					return map_get_server_hash(px, hash);
@@ -394,16 +417,15 @@ static struct server *get_server_ph_post(struct stream *s, const struct server *
  * performance by avoiding bounces between servers in contexts where sessions
  * are shared but cookies are not usable. If the parameter is not found, NULL
  * is returned. If any server is found, it will be returned. If no valid server
- * is found, NULL is returned.
+ * is found, NULL is returned. When lbprm.arg_opt1 is set, the hash will only
+ * apply to the middle part of a domain name ("use_domain_only" option).
  */
 static struct server *get_server_hh(struct stream *s, const struct server *avoid)
 {
 	unsigned int hash = 0;
-	struct http_txn *txn  = s->txn;
 	struct proxy    *px   = s->be;
-	unsigned int     plen = px->hh_len;
+	unsigned int     plen = px->lbprm.arg_len;
 	unsigned long    len;
-	struct hdr_ctx   ctx;
 	const char      *p;
 	const char *start, *end;
 
@@ -411,25 +433,46 @@ static struct server *get_server_hh(struct stream *s, const struct server *avoid
 	if (px->lbprm.tot_weight == 0)
 		return NULL;
 
-	ctx.idx = 0;
-
-	/* if the message is chunked, we skip the chunk size, but use the value as len */
-	http_find_header2(px->hh_name, plen, c_ptr(&s->req, -http_hdr_rewind(&txn->req)), &txn->hdr_idx, &ctx);
-
-	/* if the header is not found or empty, let's fallback to round robin */
-	if (!ctx.idx || !ctx.vlen)
-		return NULL;
-
 	/* note: we won't hash if there's only one server left */
 	if (px->lbprm.tot_used == 1)
 		goto hash_done;
 
-	/* Found a the hh_name in the headers.
-	 * we will compute the hash based on this value ctx.val.
-	 */
-	len = ctx.vlen;
-	p = (char *)ctx.line + ctx.val;
-	if (!px->hh_match_domain) {
+	if (!IS_HTX_STRM(s)) {
+		struct http_txn *txn = s->txn;
+		struct hdr_ctx   ctx = { .idx = 0 };
+
+		/* if the message is chunked, we skip the chunk size, but use the value as len */
+		http_find_header2(px->lbprm.arg_str, plen, c_ptr(&s->req, -http_hdr_rewind(&txn->req)),
+				  &txn->hdr_idx, &ctx);
+
+		/* if the header is not found or empty, let's fallback to round robin */
+		if (!ctx.idx || !ctx.vlen)
+			return NULL;
+
+		/* Found the param_name in the headers.
+		 * we will compute the hash based on this value ctx.val.
+		 */
+		len = ctx.vlen;
+		p = (char *)ctx.line + ctx.val;
+	}
+	else {
+		struct htx *htx = htxbuf(&s->req.buf);
+		struct http_hdr_ctx ctx = { .blk = NULL };
+
+		http_find_header(htx, ist2(px->lbprm.arg_str, plen), &ctx, 0);
+
+		/* if the header is not found or empty, let's fallback to round robin */
+		if (!ctx.blk || !ctx.value.len)
+			return NULL;
+
+		/* Found a the param_name in the headers.
+		 * we will compute the hash based on this value ctx.val.
+		 */
+		len = ctx.value.len;
+		p   = ctx.value.ptr;
+	}
+
+	if (!px->lbprm.arg_opt1) {
 		hash = gen_hash(px, p, len);
 	} else {
 		int dohash = 0;
@@ -465,7 +508,7 @@ static struct server *get_server_hh(struct stream *s, const struct server *avoid
 	if ((px->lbprm.algo & BE_LB_HASH_MOD) == BE_LB_HMOD_AVAL)
 		hash = full_hash(hash);
  hash_done:
-	if (px->lbprm.algo & BE_LB_LKUP_CHTREE)
+	if ((px->lbprm.algo & BE_LB_LKUP) == BE_LB_LKUP_CHTREE)
 		return chash_get_server_hash(px, hash, avoid);
 	else
 		return map_get_server_hash(px, hash);
@@ -490,7 +533,7 @@ static struct server *get_server_rch(struct stream *s, const struct server *avoi
 	rewind = co_data(&s->req);
 	c_rew(&s->req, rewind);
 
-	ret = fetch_rdp_cookie_name(s, &smp, px->hh_name, px->hh_len);
+	ret = fetch_rdp_cookie_name(s, &smp, px->lbprm.arg_str, px->lbprm.arg_len);
 	len = smp.data.u.str.data;
 
 	c_adv(&s->req, rewind);
@@ -502,7 +545,7 @@ static struct server *get_server_rch(struct stream *s, const struct server *avoi
 	if (px->lbprm.tot_used == 1)
 		goto hash_done;
 
-	/* Found a the hh_name in the headers.
+	/* Found the param_name in the headers.
 	 * we will compute the hash based on this value ctx.val.
 	 */
 	hash = gen_hash(px, smp.data.u.str.area, len);
@@ -510,7 +553,7 @@ static struct server *get_server_rch(struct stream *s, const struct server *avoi
 	if ((px->lbprm.algo & BE_LB_HASH_MOD) == BE_LB_HMOD_AVAL)
 		hash = full_hash(hash);
  hash_done:
-	if (px->lbprm.algo & BE_LB_LKUP_CHTREE)
+	if ((px->lbprm.algo & BE_LB_LKUP) == BE_LB_LKUP_CHTREE)
 		return chash_get_server_hash(px, hash, avoid);
 	else
 		return map_get_server_hash(px, hash);
@@ -521,14 +564,31 @@ static struct server *get_server_rnd(struct stream *s, const struct server *avoi
 {
 	unsigned int hash = 0;
 	struct proxy  *px = s->be;
+	struct server *prev, *curr;
+	int draws = px->lbprm.arg_opt1; // number of draws
 
 	/* tot_weight appears to mean srv_count */
 	if (px->lbprm.tot_weight == 0)
 		return NULL;
 
-	/* ensure all 32 bits are covered as long as RAND_MAX >= 65535 */
-	hash = ((uint64_t)random() * ((uint64_t)RAND_MAX + 1)) ^ random();
-	return chash_get_server_hash(px, hash, avoid);
+	curr = NULL;
+	do {
+		prev = curr;
+		/* ensure all 32 bits are covered as long as RAND_MAX >= 65535 */
+		hash = ((uint64_t)random() * ((uint64_t)RAND_MAX + 1)) ^ random();
+		curr = chash_get_server_hash(px, hash, avoid);
+		if (!curr)
+			break;
+
+		/* compare the new server to the previous best choice and pick
+		 * the one with the least currently served requests.
+		 */
+		if (prev && prev != curr &&
+		    curr->served * prev->cur_eweight > prev->served * curr->cur_eweight)
+			curr = prev;
+	} while (--draws > 0);
+
+	return curr;
 }
 
 /*
@@ -587,14 +647,14 @@ int assign_server(struct stream *s)
 	s->target = NULL;
 
 	if ((s->be->lbprm.algo & BE_LB_KIND) != BE_LB_KIND_HI &&
-	    ((s->txn && s->txn->flags & TX_PREFER_LAST) ||
+	    ((s->sess->flags & SESS_FL_PREFER_LAST) ||
 	     (s->be->options & PR_O_PREF_LAST))) {
 		struct sess_srv_list *srv_list;
 		list_for_each_entry(srv_list, &s->sess->srv_list, srv_list) {
 			struct server *tmpsrv = objt_server(srv_list->target);
 
 			if (tmpsrv && tmpsrv->proxy == s->be &&
-			    ((s->txn && s->txn->flags & TX_PREFER_LAST) ||
+			    ((s->sess->flags & SESS_FL_PREFER_LAST) ||
 			     (!s->be->max_ka_queue ||
 			      server_has_room(tmpsrv) || (
 			      tmpsrv->nbpend + 1 < s->be->max_ka_queue))) &&
@@ -640,7 +700,7 @@ int assign_server(struct stream *s)
 			if ((s->be->lbprm.algo & BE_LB_KIND) == BE_LB_KIND_RR) {
 				if ((s->be->lbprm.algo & BE_LB_PARM) == BE_LB_RR_RANDOM)
 					srv = get_server_rnd(s, prev_srv);
-				else if (s->be->lbprm.algo & BE_LB_LKUP_CHTREE)
+				else if ((s->be->lbprm.algo & BE_LB_LKUP) == BE_LB_LKUP_CHTREE)
 					srv = chash_get_next_server(s->be, prev_srv);
 				else
 					srv = map_get_server_rr(s->be, prev_srv);
@@ -676,9 +736,16 @@ int assign_server(struct stream *s)
 				/* URI hashing */
 				if (!s->txn || s->txn->req.msg_state < HTTP_MSG_BODY)
 					break;
-				srv = get_server_uh(s->be,
-						    c_ptr(&s->req, -http_uri_rewind(&s->txn->req)),
-						    s->txn->req.sl.rq.u_l, prev_srv);
+				if (!IS_HTX_STRM(s))
+					srv = get_server_uh(s->be,
+							    c_ptr(&s->req, -http_uri_rewind(&s->txn->req)),
+							    s->txn->req.sl.rq.u_l, prev_srv);
+				else {
+					struct ist uri;
+
+					uri = htx_sl_req_uri(http_get_stline(htxbuf(&s->req.buf)));
+					srv = get_server_uh(s->be, uri.ptr, uri.len, prev_srv);
+				}
 				break;
 
 			case BE_LB_HASH_PRM:
@@ -686,9 +753,16 @@ int assign_server(struct stream *s)
 				if (!s->txn || s->txn->req.msg_state < HTTP_MSG_BODY)
 					break;
 
-				srv = get_server_ph(s->be,
-						    c_ptr(&s->req, -http_uri_rewind(&s->txn->req)),
-						    s->txn->req.sl.rq.u_l, prev_srv);
+				if (!IS_HTX_STRM(s))
+					srv = get_server_ph(s->be,
+							    c_ptr(&s->req, -http_uri_rewind(&s->txn->req)),
+							    s->txn->req.sl.rq.u_l, prev_srv);
+				else {
+					struct ist uri;
+
+					uri = htx_sl_req_uri(http_get_stline(htxbuf(&s->req.buf)));
+					srv = get_server_ph(s->be, uri.ptr, uri.len, prev_srv);
+				}
 
 				if (!srv && s->txn->meth == HTTP_METH_POST)
 					srv = get_server_ph_post(s, prev_srv);
@@ -716,7 +790,7 @@ int assign_server(struct stream *s)
 			 * back to round robin on the map.
 			 */
 			if (!srv) {
-				if (s->be->lbprm.algo & BE_LB_LKUP_CHTREE)
+				if ((s->be->lbprm.algo & BE_LB_LKUP) == BE_LB_LKUP_CHTREE)
 					srv = chash_get_next_server(s->be, prev_srv);
 				else
 					srv = map_get_server_rr(s->be, prev_srv);
@@ -736,8 +810,8 @@ int assign_server(struct stream *s)
 			goto out;
 		}
 		else if (srv != prev_srv) {
-			HA_ATOMIC_ADD(&s->be->be_counters.cum_lbconn, 1);
-			HA_ATOMIC_ADD(&srv->counters.cum_lbconn, 1);
+			_HA_ATOMIC_ADD(&s->be->be_counters.cum_lbconn, 1);
+			_HA_ATOMIC_ADD(&srv->counters.cum_lbconn, 1);
 		}
 		s->target = &srv->obj_type;
 	}
@@ -917,11 +991,11 @@ int assign_server_and_queue(struct stream *s)
 					s->txn->flags |= TX_CK_DOWN;
 				}
 				s->flags |= SF_REDISP;
-				HA_ATOMIC_ADD(&prev_srv->counters.redispatches, 1);
-				HA_ATOMIC_ADD(&s->be->be_counters.redispatches, 1);
+				_HA_ATOMIC_ADD(&prev_srv->counters.redispatches, 1);
+				_HA_ATOMIC_ADD(&s->be->be_counters.redispatches, 1);
 			} else {
-				HA_ATOMIC_ADD(&prev_srv->counters.retries, 1);
-				HA_ATOMIC_ADD(&s->be->be_counters.retries, 1);
+				_HA_ATOMIC_ADD(&prev_srv->counters.retries, 1);
+				_HA_ATOMIC_ADD(&s->be->be_counters.retries, 1);
 			}
 		}
 	}
@@ -1001,7 +1075,12 @@ static void assign_tproxy_address(struct stream *s)
 	struct server *srv = objt_server(s->target);
 	struct conn_src *src;
 	struct connection *cli_conn;
-	struct connection *srv_conn = cs_conn(objt_cs(s->si[1].end));
+	struct connection *srv_conn;
+
+	if (objt_cs(s->si[1].end))
+		srv_conn = cs_conn(__objt_cs(s->si[1].end));
+	else
+		srv_conn = objt_conn(s->si[1].end);
 
 	if (srv && srv->conn_src.opts & CO_SRC_BIND)
 		src = &srv->conn_src;
@@ -1027,20 +1106,30 @@ static void assign_tproxy_address(struct stream *s)
 		if (src->bind_hdr_occ && s->txn) {
 			char *vptr;
 			size_t vlen;
-			int rewind;
 
 			/* bind to the IP in a header */
 			((struct sockaddr_in *)&srv_conn->addr.from)->sin_family = AF_INET;
 			((struct sockaddr_in *)&srv_conn->addr.from)->sin_port = 0;
 			((struct sockaddr_in *)&srv_conn->addr.from)->sin_addr.s_addr = 0;
+			if (!IS_HTX_STRM(s)) {
+				int rewind;
 
-			c_rew(&s->req, rewind = http_hdr_rewind(&s->txn->req));
-			if (http_get_hdr(&s->txn->req, src->bind_hdr_name, src->bind_hdr_len,
-					 &s->txn->hdr_idx, src->bind_hdr_occ, NULL, &vptr, &vlen)) {
-				((struct sockaddr_in *)&srv_conn->addr.from)->sin_addr.s_addr =
-					htonl(inetaddr_host_lim(vptr, vptr + vlen));
+				c_rew(&s->req, rewind = http_hdr_rewind(&s->txn->req));
+				if (http_get_hdr(&s->txn->req, src->bind_hdr_name, src->bind_hdr_len,
+						 &s->txn->hdr_idx, src->bind_hdr_occ, NULL, &vptr, &vlen)) {
+					((struct sockaddr_in *)&srv_conn->addr.from)->sin_addr.s_addr =
+						htonl(inetaddr_host_lim(vptr, vptr + vlen));
+				}
+				c_adv(&s->req, rewind);
+			}
+			else {
+				if (http_get_htx_hdr(htxbuf(&s->req.buf),
+						     ist2(src->bind_hdr_name, src->bind_hdr_len),
+						     src->bind_hdr_occ, NULL, &vptr, &vlen)) {
+					((struct sockaddr_in *)&srv_conn->addr.from)->sin_addr.s_addr =
+						htonl(inetaddr_host_lim(vptr, vptr + vlen));
+				}
 			}
-			c_adv(&s->req, rewind);
 		}
 		break;
 	default:
@@ -1083,6 +1172,8 @@ static int conn_complete_server(struct connection *conn)
 	return 0;
 
 fail:
+	si_detach_endpoint(&s->si[1]);
+
 	if (cs)
 		cs_free(cs);
 	/* kill the connection now */
@@ -1121,6 +1212,7 @@ int connect_server(struct stream *s)
 	int reuse = 0;
 	int reuse_orphan = 0;
 	int init_mux = 0;
+	int alloced_cs = 0;
 	int err;
 
 
@@ -1129,14 +1221,25 @@ int connect_server(struct stream *s)
 	 * to use it.
 	 */
 	srv_cs = objt_cs(s->si[1].end);
-	if (srv_cs) {
-		old_conn = srv_conn = cs_conn(srv_cs);
-		if (old_conn && (!old_conn->target || old_conn->target == s->target)) {
-			old_conn->flags &= ~(CO_FL_ERROR | CO_FL_SOCK_RD_SH | CO_FL_SOCK_WR_SH);
-			srv_cs->flags &= ~(CS_FL_ERROR | CS_FL_EOS | CS_FL_REOS);
+	if (!srv_cs)
+		srv_conn = objt_conn(s->si[1].end);
+	else
+		srv_conn = cs_conn(srv_cs);
+
+	if (srv_conn) {
+		if (!srv_conn->target || srv_conn->target == s->target) {
+			srv_conn->flags &= ~(CO_FL_ERROR | CO_FL_SOCK_RD_SH | CO_FL_SOCK_WR_SH);
+			if (srv_cs)
+				srv_cs->flags &= ~(CS_FL_ERROR | CS_FL_EOS);
 			reuse = 1;
+			old_conn = srv_conn;
+		} else {
+			srv_conn = NULL;
+			si_release_endpoint(&s->si[1]);
 		}
-	} else {
+	}
+
+	if (!old_conn) {
 		struct sess_srv_list *srv_list;
 		list_for_each_entry(srv_list, &s->sess->srv_list, srv_list) {
 			if (srv_list->target == s->target) {
@@ -1201,9 +1304,10 @@ int connect_server(struct stream *s)
 		    (((s->be->options & PR_O_REUSE_MASK) == PR_O_REUSE_ALWS) ||
 		    (((s->be->options & PR_O_REUSE_MASK) != PR_O_REUSE_NEVR) &&
 		     s->txn && (s->txn->flags & TX_NOT_FIRST)))) {
-			srv_conn = LIST_ELEM(srv->idle_orphan_conns[tid].n,
-			    struct connection *, list);
-			reuse_orphan = 1;
+			srv_conn = LIST_POP_LOCKED(&srv->idle_orphan_conns[tid],
+			                           struct connection *, list);
+			if (srv_conn)
+				reuse_orphan = 1;
 		}
 
 		/* If we've picked a connection from the pool, we now have to
@@ -1231,16 +1335,65 @@ int connect_server(struct stream *s)
 				reuse = 0;
 		}
 	}
+	if (((!reuse || (srv_conn && !(srv_conn->flags & CO_FL_CONNECTED)))
+	    && ha_used_fds > global.tune.pool_high_count) && srv->idle_orphan_conns) {
+		struct connection *tokill_conn;
+
+		/* We can't reuse a connection, and e have more FDs than deemd
+		 * acceptable, attempt to kill an idling connection
+		 */
+		/* First, try from our own idle list */
+		tokill_conn = LIST_POP_LOCKED(&srv->idle_orphan_conns[tid],
+		    struct connection *, list);
+		if (tokill_conn)
+			tokill_conn->mux->destroy(tokill_conn->ctx);
+		/* If not, iterate over other thread's idling pool, and try to grab one */
+		else {
+			int i;
+
+			for (i = 0; i < global.nbthread; i++) {
+				if (i == tid)
+					continue;
+
+				// just silence stupid gcc which reports an absurd
+				// out-of-bounds warning for <i> which is always
+				// exactly zero without threads, but it seems to
+				// see it possibly larger.
+				ALREADY_CHECKED(i);
+
+				tokill_conn = LIST_POP_LOCKED(&srv->idle_orphan_conns[i],
+				    struct connection *, list);
+				if (tokill_conn) {
+					/* We got one, put it into the concerned thread's to kill list, and wake it's kill task */
+
+					LIST_ADDQ_LOCKED(&toremove_connections[i],
+					    &tokill_conn->list);
+					task_wakeup(idle_conn_cleanup[i], TASK_WOKEN_OTHER);
+					break;
+				}
+			}
+		}
+
+	}
 	/* If we're really reusing the connection, remove it from the orphan
 	 * list and add it back to the idle list.
 	 */
-	if (reuse && reuse_orphan) {
-		LIST_DEL(&srv_conn->list);
-		srv_conn->idle_time = 0;
-		srv->curr_idle_conns--;
-		LIST_ADDQ(&srv->idle_conns[tid], &srv_conn->list);
-		if (LIST_ISEMPTY(&srv->idle_orphan_conns[tid]))
-			task_unlink_wq(srv->idle_task[tid]);
+	if (reuse) {
+		if (reuse_orphan) {
+			srv_conn->idle_time = 0;
+			_HA_ATOMIC_SUB(&srv->curr_idle_conns, 1);
+			__ha_barrier_atomic_store();
+			srv->curr_idle_thr[tid]--;
+			LIST_ADDQ(&srv->idle_conns[tid], &srv_conn->list);
+		}
+		else {
+			if (srv_conn->flags & CO_FL_SESS_IDLE) {
+				struct session *sess = srv_conn->owner;
+
+				srv_conn->flags &= ~CO_FL_SESS_IDLE;
+				sess->idle_conns--;
+			}
+		}
 	}
 
 	/* We're about to use another connection, let the mux know we're
@@ -1252,49 +1405,70 @@ int connect_server(struct stream *s)
 		if (sess) {
 			if (old_conn && !(old_conn->flags & CO_FL_PRIVATE) &&
 			    old_conn->mux != NULL) {
+				if (old_conn->flags & CO_FL_SESS_IDLE)
+					s->sess->idle_conns--;
 				session_unown_conn(s->sess, old_conn);
 				old_conn->owner = sess;
-				if (!session_add_conn(sess, old_conn, s->target)) {
+				if (!session_add_conn(sess, old_conn, old_conn->target)) {
+					old_conn->flags &= ~CO_FL_SESS_IDLE;
 					old_conn->owner = NULL;
-					old_conn->mux->destroy(old_conn);
+					old_conn->mux->destroy(old_conn->ctx);
 				} else
 					session_check_idle_conn(sess, old_conn);
 			}
 		}
 	}
 
-	if (!reuse) {
-		srv_conn = conn_new();
-		srv_cs = NULL;
-	} else {
+	if (reuse) {
 		/* We already created a cs earlier when using http_proxy, so
 		 * only create a new one if we don't have one already.
 		 */
-		if (!srv_cs) {
-			if (srv_conn->mux->avail_streams(srv_conn) == 1) {
+		if (!srv_cs && srv_conn->mux) {
+			int avail = srv_conn->mux->avail_streams(srv_conn);
+
+			if (avail <= 1) {
 				/* No more streams available, remove it from the list */
 				LIST_DEL(&srv_conn->list);
 				LIST_INIT(&srv_conn->list);
 			}
-			srv_cs = srv_conn->mux->attach(srv_conn, s->sess);
-			if (srv_cs)
-				si_attach_cs(&s->si[1], srv_cs);
+
+			if (avail >= 1) {
+				srv_cs = srv_conn->mux->attach(srv_conn, s->sess);
+				if (srv_cs) {
+					alloced_cs = 1;
+					si_attach_cs(&s->si[1], srv_cs);
+				} else
+					srv_conn = NULL;
+			}
 			else
 				srv_conn = NULL;
 		}
+		/* otherwise srv_conn is left intact */
 	}
+	else
+		srv_conn = NULL;
+
+	/* no reuse or failed to reuse the connection above, pick a new one */
+	if (!srv_conn) {
+		srv_conn = conn_new();
+		if (srv_conn)
+			srv_conn->target = s->target;
+		srv_cs = NULL;
+	}
+
 	if (srv_conn && old_conn != srv_conn) {
 		if (srv_conn->owner)
 			session_unown_conn(srv_conn->owner, srv_conn);
 		srv_conn->owner = s->sess;
-		if (!session_add_conn(s->sess, srv_conn, s->target)) {
+		if (!session_add_conn(s->sess, srv_conn, srv_conn->target)) {
 			/* If we failed to attach the connection, detach the
 			 * conn_stream, possibly destroying the connection */
-			cs_destroy(srv_cs);
+			if (alloced_cs)
+				si_release_endpoint(&s->si[1]);
 			srv_conn->owner = NULL;
-			if (!srv_add_to_idle_list(objt_server(srv_conn->target), srv_conn))
+			if (srv_conn->mux && !srv_add_to_idle_list(objt_server(srv_conn->target), srv_conn))
 			/* The server doesn't want it, let's kill the connection right away */
-				srv_conn->mux->destroy(srv_conn);
+				srv_conn->mux->destroy(srv_conn->ctx);
 			srv_conn = NULL;
 
 		}
@@ -1310,9 +1484,6 @@ int connect_server(struct stream *s)
 	}
 
 	if (!conn_xprt_ready(srv_conn) && !srv_conn->mux) {
-		/* the target was only on the stream, assign it to the SI now */
-		srv_conn->target = s->target;
-
 		/* set the correct protocol on the output stream interface */
 		if (srv)
 			conn_prepare(srv_conn, protocol_by_family(srv_conn->addr.to.ss_family), srv->xprt);
@@ -1328,7 +1499,7 @@ int connect_server(struct stream *s)
 #if defined(USE_OPENSSL) && defined(TLSEXT_TYPE_application_layer_protocol_negotiation)
 		if (!srv ||
 		    ((!(srv->ssl_ctx.alpn_str) && !(srv->ssl_ctx.npn_str)) ||
-		    srv->mux_proto))
+		    srv->mux_proto || s->be->mode != PR_MODE_HTTP))
 #endif
 		{
 			srv_cs = objt_cs(s->si[1].end);
@@ -1362,12 +1533,18 @@ int connect_server(struct stream *s)
 
 		if (srv && srv->pp_opts) {
 			srv_conn->flags |= CO_FL_PRIVATE;
+			srv_conn->flags |= CO_FL_SEND_PROXY;
 			srv_conn->send_proxy_ofs = 1; /* must compute size */
 			if (cli_conn)
 				conn_get_to_addr(cli_conn);
 		}
 
 		assign_tproxy_address(s);
+
+		if (srv && (srv->flags & SRV_F_SOCKS4_PROXY)) {
+			srv_conn->send_proxy_ofs = 1;
+			srv_conn->flags |= CO_FL_SOCKS4;
+		}
 	}
 	else if (!conn_xprt_ready(srv_conn)) {
 		if (srv_conn->mux->reset)
@@ -1385,13 +1562,13 @@ int connect_server(struct stream *s)
 		s->si[1].flags |= SI_FL_NOLINGER;
 
 	if (s->flags & SF_SRV_REUSED) {
-		HA_ATOMIC_ADD(&s->be->be_counters.reuse, 1);
+		_HA_ATOMIC_ADD(&s->be->be_counters.reuse, 1);
 		if (srv)
-			HA_ATOMIC_ADD(&srv->counters.reuse, 1);
+			_HA_ATOMIC_ADD(&srv->counters.reuse, 1);
 	} else {
-		HA_ATOMIC_ADD(&s->be->be_counters.connect, 1);
+		_HA_ATOMIC_ADD(&s->be->be_counters.connect, 1);
 		if (srv)
-			HA_ATOMIC_ADD(&srv->counters.connect, 1);
+			_HA_ATOMIC_ADD(&srv->counters.connect, 1);
 	}
 
 	err = si_connect(&s->si[1], srv_conn);
@@ -1401,8 +1578,10 @@ int connect_server(struct stream *s)
 	 * fail, and flag the connection as CO_FL_ERROR.
 	 */
 	if (init_mux) {
-		if (conn_install_mux_be(srv_conn, srv_cs, s->sess) < 0)
+		if (conn_install_mux_be(srv_conn, srv_cs, s->sess) < 0) {
+			conn_full_close(srv_conn);
 			return SF_ERR_INTERNAL;
+		}
 		/* If we're doing http-reuse always, and the connection
 		 * is an http2 connection, add it to the available list,
 		 * so that others can use it right away.
@@ -1411,17 +1590,32 @@ int connect_server(struct stream *s)
 		    srv_conn->mux->avail_streams(srv_conn) > 0)
 			LIST_ADD(&srv->idle_conns[tid], &srv_conn->list);
 	}
-
-
-#ifdef USE_OPENSSL
-	if (!reuse && cli_conn && srv &&
-	    (srv->ssl_ctx.options & SRV_SSL_O_EARLY_DATA) &&
-		    (cli_conn->flags & CO_FL_EARLY_DATA) &&
-		    !channel_is_empty(si_oc(&s->si[1])) &&
-		    srv_conn->flags & CO_FL_SSL_WAIT_HS) {
-		srv_conn->flags &= ~(CO_FL_SSL_WAIT_HS | CO_FL_WAIT_L6_CONN);
-		srv_conn->flags |= CO_FL_EARLY_SSL_HS;
+	/* The CO_FL_SEND_PROXY flag may have been set by the connect method,
+	 * if so, add our handshake pseudo-XPRT now.
+	 */
+	if ((srv_conn->flags & CO_FL_HANDSHAKE_NOSSL)) {
+		if (xprt_add_hs(srv_conn) < 0) {
+			conn_full_close(srv_conn);
+			return SF_ERR_INTERNAL;
+		}
 	}
+
+
+#if USE_OPENSSL && (defined(OPENSSL_IS_BORINGSSL) || (HA_OPENSSL_VERSION_NUMBER >= 0x10101000L))
+
+	if (!reuse && cli_conn && srv && srv_conn->mux &&
+	    (srv->ssl_ctx.options & SRV_SSL_O_EARLY_DATA) &&
+	    /* Only attempt to use early data if either the client sent
+	     * early data, so that we know it can handle a 425, or if
+	     * we are allwoed to retry requests on early data failure, and
+	     * it's our first try
+	     */
+	    ((cli_conn->flags & CO_FL_EARLY_DATA) ||
+	     ((s->be->retry_type & PR_RE_EARLY_ERROR) &&
+	      s->si[1].conn_retries == s->be->conn_retries)) &&
+	    !channel_is_empty(si_oc(&s->si[1])) &&
+	    srv_conn->flags & CO_FL_SSL_WAIT_HS)
+		srv_conn->flags &= ~(CO_FL_SSL_WAIT_HS | CO_FL_WAIT_L6_CONN);
 #endif
 
 	if (err != SF_ERR_NONE)
@@ -1434,7 +1628,7 @@ int connect_server(struct stream *s)
 		int count;
 
 		s->flags |= SF_CURR_SESS;
-		count = HA_ATOMIC_ADD(&srv->cur_sess, 1);
+		count = _HA_ATOMIC_ADD(&srv->cur_sess, 1);
 		HA_ATOMIC_UPDATE_MAX(&srv->counters.cur_sess_max, count);
 		if (s->be->lbprm.server_take_conn)
 			s->be->lbprm.server_take_conn(srv);
@@ -1444,18 +1638,32 @@ int connect_server(struct stream *s)
 			struct sample *smp;
 			int rewind;
 
-			/* Tricky case : we have already scheduled the pending
-			 * HTTP request or TCP data for leaving. So in HTTP we
-			 * rewind exactly the headers, otherwise we rewind the
-			 * output data.
-			 */
-			rewind = s->txn ? http_hdr_rewind(&s->txn->req) : co_data(&s->req);
-			c_rew(&s->req, rewind);
+			if (!IS_HTX_STRM(s)) {
+				/* Tricky case : we have already scheduled the pending
+				 * HTTP request or TCP data for leaving. So in HTTP we
+				 * rewind exactly the headers, otherwise we rewind the
+				 * output data.
+				 */
+				rewind = s->txn ? http_hdr_rewind(&s->txn->req) : co_data(&s->req);
+				c_rew(&s->req, rewind);
 
-			smp = sample_fetch_as_type(s->be, s->sess, s, SMP_OPT_DIR_REQ | SMP_OPT_FINAL, srv->ssl_ctx.sni, SMP_T_STR);
+				smp = sample_fetch_as_type(s->be, s->sess, s, SMP_OPT_DIR_REQ | SMP_OPT_FINAL,
+							   srv->ssl_ctx.sni, SMP_T_STR);
 
-			/* restore the pointers */
-			c_adv(&s->req, rewind);
+				/* restore the pointers */
+				c_adv(&s->req, rewind);
+			}
+			else {
+				/* rewind the output data. */
+				rewind = co_data(&s->req);
+				c_rew(&s->req, rewind);
+
+				smp = sample_fetch_as_type(s->be, s->sess, s, SMP_OPT_DIR_REQ | SMP_OPT_FINAL,
+							   srv->ssl_ctx.sni, SMP_T_STR);
+
+				/* restore the pointers */
+				c_adv(&s->req, rewind);
+			}
 
 			if (smp_make_safe(smp)) {
 				ssl_sock_set_servername(srv_conn,
@@ -1513,8 +1721,8 @@ int srv_redispatch_connect(struct stream *s)
 			s->si[1].err_type = SI_ET_QUEUE_ERR;
 		}
 
-		HA_ATOMIC_ADD(&srv->counters.failed_conns, 1);
-		HA_ATOMIC_ADD(&s->be->be_counters.failed_conns, 1);
+		_HA_ATOMIC_ADD(&srv->counters.failed_conns, 1);
+		_HA_ATOMIC_ADD(&s->be->be_counters.failed_conns, 1);
 		return 1;
 
 	case SRV_STATUS_NOSRV:
@@ -1523,7 +1731,7 @@ int srv_redispatch_connect(struct stream *s)
 			s->si[1].err_type = SI_ET_CONN_ERR;
 		}
 
-		HA_ATOMIC_ADD(&s->be->be_counters.failed_conns, 1);
+		_HA_ATOMIC_ADD(&s->be->be_counters.failed_conns, 1);
 		return 1;
 
 	case SRV_STATUS_QUEUED:
@@ -1543,8 +1751,8 @@ int srv_redispatch_connect(struct stream *s)
 		if (srv)
 			srv_set_sess_last(srv);
 		if (srv)
-			HA_ATOMIC_ADD(&srv->counters.failed_conns, 1);
-		HA_ATOMIC_ADD(&s->be->be_counters.failed_conns, 1);
+			_HA_ATOMIC_ADD(&srv->counters.failed_conns, 1);
+		_HA_ATOMIC_ADD(&s->be->be_counters.failed_conns, 1);
 
 		/* release other streams waiting for this server */
 		if (may_dequeue_tasks(srv, s->be))
@@ -1563,7 +1771,7 @@ int srv_redispatch_connect(struct stream *s)
 void set_backend_down(struct proxy *be)
 {
 	be->last_change = now.tv_sec;
-	HA_ATOMIC_ADD(&be->down_trans, 1);
+	_HA_ATOMIC_ADD(&be->down_trans, 1);
 
 	if (!(global.mode & MODE_STARTING)) {
 		ha_alert("%s '%s' has no server available!\n", proxy_type_str(be), be->id);
@@ -1710,9 +1918,31 @@ int backend_parse_balance(const char **args, char **err, struct proxy *curproxy)
 		curproxy->lbprm.algo &= ~BE_LB_ALGO;
 		curproxy->lbprm.algo |= BE_LB_ALGO_LC;
 	}
-	else if (!strcmp(args[0], "random")) {
+	else if (!strncmp(args[0], "random", 6)) {
 		curproxy->lbprm.algo &= ~BE_LB_ALGO;
 		curproxy->lbprm.algo |= BE_LB_ALGO_RND;
+		curproxy->lbprm.arg_opt1 = 2;
+
+		if (*(args[0] + 6) == '(' && *(args[0] + 7) != ')') { /* number of draws */
+			const char *beg;
+			char *end;
+
+			beg = args[0] + 7;
+			curproxy->lbprm.arg_opt1 = strtol(beg, &end, 0);
+
+			if (*end != ')') {
+				if (!*end)
+					memprintf(err, "random : missing closing parenthesis.");
+				else
+					memprintf(err, "random : unexpected character '%c' after argument.", *end);
+				return -1;
+			}
+
+			if (curproxy->lbprm.arg_opt1 < 1) {
+				memprintf(err, "random : number of draws must be at least 1.");
+				return -1;
+			}
+		}
 	}
 	else if (!strcmp(args[0], "source")) {
 		curproxy->lbprm.algo &= ~BE_LB_ALGO;
@@ -1723,8 +1953,9 @@ int backend_parse_balance(const char **args, char **err, struct proxy *curproxy)
 
 		curproxy->lbprm.algo &= ~BE_LB_ALGO;
 		curproxy->lbprm.algo |= BE_LB_ALGO_UH;
-
-		curproxy->uri_whole = 0;
+		curproxy->lbprm.arg_opt1 = 0; // "whole"
+		curproxy->lbprm.arg_opt2 = 0; // "len"
+		curproxy->lbprm.arg_opt3 = 0; // "depth"
 
 		while (*args[arg]) {
 			if (!strcmp(args[arg], "len")) {
@@ -1732,7 +1963,7 @@ int backend_parse_balance(const char **args, char **err, struct proxy *curproxy)
 					memprintf(err, "%s : '%s' expects a positive integer (got '%s').", args[0], args[arg], args[arg+1]);
 					return -1;
 				}
-				curproxy->uri_len_limit = atoi(args[arg+1]);
+				curproxy->lbprm.arg_opt2 = atoi(args[arg+1]);
 				arg += 2;
 			}
 			else if (!strcmp(args[arg], "depth")) {
@@ -1743,11 +1974,11 @@ int backend_parse_balance(const char **args, char **err, struct proxy *curproxy)
 				/* hint: we store the position of the ending '/' (depth+1) so
 				 * that we avoid a comparison while computing the hash.
 				 */
-				curproxy->uri_dirs_depth1 = atoi(args[arg+1]) + 1;
+				curproxy->lbprm.arg_opt3 = atoi(args[arg+1]) + 1;
 				arg += 2;
 			}
 			else if (!strcmp(args[arg], "whole")) {
-				curproxy->uri_whole = 1;
+				curproxy->lbprm.arg_opt1 = 1;
 				arg += 1;
 			}
 			else {
@@ -1764,9 +1995,9 @@ int backend_parse_balance(const char **args, char **err, struct proxy *curproxy)
 		curproxy->lbprm.algo &= ~BE_LB_ALGO;
 		curproxy->lbprm.algo |= BE_LB_ALGO_PH;
 
-		free(curproxy->url_param_name);
-		curproxy->url_param_name = strdup(args[1]);
-		curproxy->url_param_len  = strlen(args[1]);
+		free(curproxy->lbprm.arg_str);
+		curproxy->lbprm.arg_str = strdup(args[1]);
+		curproxy->lbprm.arg_len = strlen(args[1]);
 		if (*args[2]) {
 			if (strcmp(args[2], "check_post")) {
 				memprintf(err, "%s only accepts 'check_post' modifier (got '%s').", args[0], args[2]);
@@ -1788,17 +2019,17 @@ int backend_parse_balance(const char **args, char **err, struct proxy *curproxy)
 		curproxy->lbprm.algo &= ~BE_LB_ALGO;
 		curproxy->lbprm.algo |= BE_LB_ALGO_HH;
 
-		free(curproxy->hh_name);
-		curproxy->hh_len  = end - beg;
-		curproxy->hh_name = my_strndup(beg, end - beg);
-		curproxy->hh_match_domain = 0;
+		free(curproxy->lbprm.arg_str);
+		curproxy->lbprm.arg_len = end - beg;
+		curproxy->lbprm.arg_str = my_strndup(beg, end - beg);
+		curproxy->lbprm.arg_opt1 = 0;
 
 		if (*args[1]) {
 			if (strcmp(args[1], "use_domain_only")) {
 				memprintf(err, "%s only accepts 'use_domain_only' modifier (got '%s').", args[0], args[1]);
 				return -1;
 			}
-			curproxy->hh_match_domain = 1;
+			curproxy->lbprm.arg_opt1 = 1;
 		}
 	}
 	else if (!strncmp(args[0], "rdp-cookie", 10)) {
@@ -1816,14 +2047,14 @@ int backend_parse_balance(const char **args, char **err, struct proxy *curproxy)
 				return -1;
 			}
 
-			free(curproxy->hh_name);
-			curproxy->hh_name = my_strndup(beg, end - beg);
-			curproxy->hh_len  = end - beg;
+			free(curproxy->lbprm.arg_str);
+			curproxy->lbprm.arg_str = my_strndup(beg, end - beg);
+			curproxy->lbprm.arg_len = end - beg;
 		}
 		else if ( *(args[0] + 10 ) == '\0' ) { /* default cookie name 'mstshash' */
-			free(curproxy->hh_name);
-			curproxy->hh_name = strdup("mstshash");
-			curproxy->hh_len  = strlen(curproxy->hh_name);
+			free(curproxy->lbprm.arg_str);
+			curproxy->lbprm.arg_str = strdup("mstshash");
+			curproxy->lbprm.arg_len = strlen(curproxy->lbprm.arg_str);
 		}
 		else { /* syntax */
 			memprintf(err, "rdp-cookie : missing cookie name.");
