@@ -28,7 +28,7 @@ struct htx_blk *htx_defrag(struct htx *htx, struct htx_blk *blk)
 	struct htx_blk *newblk, *oldblk;
 	uint32_t new, old, blkpos;
 	uint32_t addr, blksz;
-	int32_t sl_off = -1;
+	int32_t first = -1;
 
 	if (!htx->used)
 		return NULL;
@@ -42,19 +42,17 @@ struct htx_blk *htx_defrag(struct htx *htx, struct htx_blk *blk)
 	/* start from the head */
 	for (old = htx_get_head(htx); old != -1; old = htx_get_next(htx, old)) {
 		oldblk = htx_get_blk(htx, old);
-		if (htx_get_blk_type(oldblk) == HTX_BLK_UNUSED) {
-			htx->used--;
+		if (htx_get_blk_type(oldblk) == HTX_BLK_UNUSED)
 			continue;
-		}
 
 		newblk = htx_get_blk(tmp, new);
 		newblk->addr = addr;
 		newblk->info = oldblk->info;
 		blksz = htx_get_blksz(oldblk);
 
-		/* update the start-line offset */
-		if (htx->sl_off == oldblk->addr)
-			sl_off = addr;
+		/* update the start-line position */
+		if (htx->first == old)
+			first = new;
 
 		/* if <blk> is defined, set its new position */
 		if (blk != NULL && blk == oldblk)
@@ -64,16 +62,47 @@ struct htx_blk *htx_defrag(struct htx *htx, struct htx_blk *blk)
 		new++;
 		addr += blksz;
 
-	} while (new < htx->used);
+	}
 
-	htx->sl_off = sl_off;
-	htx->wrap = htx->used;
-	htx->front = htx->tail = new - 1;
+	htx->used = new;
+	htx->first = first;
+	htx->head = 0;
+	htx->tail = new - 1;
+	htx->head_addr = htx->end_addr = 0;
+	htx->tail_addr = addr;
 	memcpy((void *)htx->blocks, (void *)tmp->blocks, htx->size);
 
 	return ((blkpos == -1) ? NULL : htx_get_blk(htx, blkpos));
 }
 
+static void htx_defrag_blks(struct htx *htx)
+{
+	int32_t pos, new;
+
+	new = 0;
+	for (pos = htx_get_head(htx); pos != -1; pos = htx_get_next(htx, pos)) {
+		struct htx_blk *posblk, *newblk;
+
+		if (pos == new) {
+			new++;
+			continue;
+		}
+
+		posblk = htx_get_blk(htx, pos);
+		if (htx_get_blk_type(posblk) == HTX_BLK_UNUSED)
+			continue;
+
+		if (htx->first == pos)
+			htx->first = new;
+		newblk = htx_get_blk(htx, new++);
+		newblk->info = posblk->info;
+		newblk->addr = posblk->addr;
+	}
+	BUG_ON(!new);
+	htx->head = 0;
+	htx->tail = new - 1;
+}
+
 /* Reserves a new block in the HTTP message <htx> with a content of <blksz>
  * bytes. If there is not enough space, NULL is returned. Otherwise the reserved
  * block is returned and the HTTP message is updated. Space for this new block
@@ -82,118 +111,194 @@ struct htx_blk *htx_defrag(struct htx *htx, struct htx_blk *blk)
  */
 static struct htx_blk *htx_reserve_nxblk(struct htx *htx, uint32_t blksz)
 {
-	struct htx_blk *blk, *prevblk, *headblk, *frtblk;
-	uint32_t used;
-	uint32_t tail;
-	uint32_t prev;
-	uint32_t wrap;
-	uint32_t head;
-	int32_t headroom, tailroom;
+	struct htx_blk *blk;
+	uint32_t used, tail;
+	uint32_t headroom, tailroom;
 
 	if (blksz > htx_free_data_space(htx))
 		return NULL; /* full */
 
 	if (!htx->used) {
 		/* Empty message */
-		htx->front = htx->tail = 0;
-		htx->wrap  = htx->used = 1;
+		htx->head = htx->tail = htx->first = 0;
+		htx->used = 1;
 		blk = htx_get_blk(htx, htx->tail);
 		blk->addr = 0;
 		htx->data = blksz;
+		htx->tail_addr = blksz;
 		return blk;
 	}
 
-	used = htx->used + 1;
-	tail = htx->tail + 1;
-	prev = htx->tail;
-	wrap = htx->wrap;
-	head = htx_get_head(htx);
-
-	if (tail == wrap) {
-		frtblk = htx_get_blk(htx, htx->front);
-
-		/* Blocks don't wrap for now. We either need to push the new one
-		 * on top of others or to defragement the table. */
-		if (sizeof(htx->blocks[0]) * htx_pos_to_idx(htx, wrap+1) >= frtblk->addr + htx_get_blksz(frtblk))
-			wrap++;
-		else if (tail >= used) /* There is hole at the beginning */
-			tail = 0;
-		else {
-			/* No more room, tail hits data. We have to realign the
-			 * whole message. */
-			goto defrag;
-		}
-	}
-	else if (used >= wrap) {
-		/* We have hit the tail, we need to reorganize the blocks. */
-		goto defrag;
-	}
-
-	/* Now we have updated tail, used and wrap, we know that there is some
-	 * available room at least from the protocol's perspective. This space
-	 * is split in two areas :
-	 *
-	 *   1: the space between the beginning of the blocks table and the
-	 *      front data's address. This space will only be used if data don't
-	 *      wrap yet.
-
-	 *   2: If the previous tail was the front block, the space between the
-	 *      beginning of the message and the head data's address.
-	 *      Otherwise, the space between the tail data's address and the
-	 *      tail's one.
+	/* Find the block's position. First, we try to get the next position in
+	 * the message, increasing the tail by one. If this position is not
+	 * available with some holes, we try to defrag the blocks without
+	 * touching their paylood. If it is impossible, we fully defrag the
+	 * message.
 	 */
-	prevblk = htx_get_blk(htx, prev);
-	headblk = htx_get_blk(htx, head);
-	if (prevblk->addr >= headblk->addr) {
-		/* the area was contiguous */
-		frtblk = htx_get_blk(htx, htx->front);
-		tailroom = sizeof(htx->blocks[0]) * htx_pos_to_idx(htx, wrap) - (frtblk->addr + htx_get_blksz(frtblk));
-		headroom = headblk->addr;
+	tail = htx->tail + 1;
+	if (sizeof(htx->blocks[0]) * htx_pos_to_idx(htx, tail) >= htx->tail_addr)
+		used = htx->used + 1;
+	else if (tail > htx->used) {
+		htx_defrag_blks(htx);
+		tail = htx->tail + 1;
+		used = htx->used + 1;
+		BUG_ON(sizeof(htx->blocks[0]) * htx_pos_to_idx(htx, tail) < htx->tail_addr);
+	}
+	else
+		goto defrag;
 
-		if (tailroom >= (int32_t)blksz) {
-			/* install upfront and update ->front */
-			blk = htx_get_blk(htx, tail);
-			blk->addr = frtblk->addr + htx_get_blksz(frtblk);
-			htx->front = tail;
-		}
-		else if (headroom >= (int32_t)blksz) {
-			blk = htx_get_blk(htx, tail);
-			blk->addr = 0;
-		}
-		else {
-			/* need to defragment the table before inserting upfront */
-			goto defrag;
-		}
+	/* Now, we have found the block's position. Try do find where to put its
+	 * payload. The free space is split in two areas:
+	 *
+	 *   * The free space in front of the blocks table. This one is used iff
+	 *     the other one was not used yet.
+	 *
+	 *   * The free space at the beginning of the message. Once this one is
+         *     used, the other one is never used again, until the next defrag.
+	 */
+	headroom = (htx->end_addr - htx->head_addr);
+	tailroom = (!htx->head_addr
+		    ? sizeof(htx->blocks[0]) * htx_pos_to_idx(htx, tail) - htx->tail_addr
+		    : 0);
+	BUG_ON((int32_t)headroom < 0);
+	BUG_ON((int32_t)tailroom < 0);
+
+	if (blksz <= tailroom) {
+		blk = htx_get_blk(htx, tail);
+		blk->addr = htx->tail_addr;
+		htx->tail_addr += blksz;
+	}
+	else if (blksz <= headroom) {
+		blk = htx_get_blk(htx, tail);
+		blk->addr = htx->head_addr;
+		htx->head_addr += blksz;
 	}
 	else {
-		/* it's already wrapped so we can't store anything in the tailroom */
-		headroom = headblk->addr - (prevblk->addr + htx_get_blksz(prevblk));
-
-		if (headroom >= (int32_t)blksz) {
-			blk = htx_get_blk(htx, tail);
-			blk->addr = prevblk->addr + htx_get_blksz(prevblk);
-		}
-		else {
-		  defrag:
-			/* need to defragment the table before inserting upfront */
-			htx_defrag(htx, NULL);
-			frtblk = htx_get_blk(htx, htx->front);
-			wrap = htx->wrap + 1;
-			tail = htx->tail + 1;
-			used = htx->used + 1;
-			blk = htx_get_blk(htx, tail);
-			blk->addr = frtblk->addr + htx_get_blksz(frtblk);
-			htx->front = tail;
-		}
+	  defrag:
+		/* need to defragment the table before inserting upfront */
+		htx_defrag(htx, NULL);
+		tail = htx->tail + 1;
+		used = htx->used + 1;
+		blk = htx_get_blk(htx, tail);
+		blk->addr = htx->tail_addr;
+		htx->tail_addr += blksz;
 	}
 
-	htx->wrap  = wrap;
 	htx->tail  = tail;
 	htx->used  = used;
 	htx->data += blksz;
+	/* Set first position if not already set */
+	if (htx->first == -1)
+		htx->first = tail;
+
+	BUG_ON((int32_t)htx->tail_addr < 0);
+	BUG_ON((int32_t)htx->head_addr < 0);
+	BUG_ON(htx->end_addr > htx->tail_addr);
+	BUG_ON(htx->head_addr > htx->end_addr);
+
 	return blk;
 }
 
+/* Prepares the block to an expansion of its payload. The payload will be
+ * expanded by <delta> bytes and we need find where this expansion will be
+ * performed. It can be a compression if <delta> is negative. This function only
+ * updates all addresses. The caller have the responsibility to performe the
+ * expansion and update the block and the HTX message accordingly. No error msut
+ * occurre. It returns following values:
+ *
+ *  0: The expansion cannot be performed, there is not enough space.
+ *
+ *  1: the expansion must be performed in place, there is enought space after
+ *      the block's payload to handle it. This is especially true if it is a
+ *      compression and not an expension.
+ *
+ *  2: the block's payload must be moved at the new block address before doing
+ *     the expansion.
+ *
+ *  3: the HTX message message must be defragmented
+ */
+static int htx_prepare_blk_expansion(struct htx *htx, struct htx_blk *blk, int32_t delta)
+{
+	uint32_t sz, tailroom, headroom;
+	int ret = 3;
+
+	headroom = (htx->end_addr - htx->head_addr);
+	tailroom = sizeof(htx->blocks[0]) * htx_pos_to_idx(htx, htx->tail) - htx->tail_addr;
+	BUG_ON((int32_t)headroom < 0);
+	BUG_ON((int32_t)tailroom < 0);
+
+	sz = htx_get_blksz(blk);
+	if (delta <= 0) {
+		/* It is a compression, it can be performed in place */
+		if (blk->addr+sz == htx->tail_addr)
+			htx->tail_addr += delta;
+		else if (blk->addr+sz == htx->head_addr)
+			htx->head_addr += delta;
+		ret = 1;
+	}
+	else if (delta > htx_free_space(htx)) {
+		/* There is not enought space to handle the expansion */
+		ret = 0;
+	}
+	else if (blk->addr+sz == htx->tail_addr) {
+		/* The block's payload is just before the tail room */
+		if (delta < tailroom) {
+			/* Expand the block's payload */
+			htx->tail_addr += delta;
+			ret = 1;
+		}
+		else if ((sz + delta) < headroom) {
+			/* Move the block's payload into the headroom */
+			blk->addr = htx->head_addr;
+			htx->tail_addr -= sz;
+			htx->head_addr += sz + delta;
+			if (blk->addr == htx->end_addr) {
+				if (htx->end_addr == htx->tail_addr) {
+					htx->tail_addr = htx->head_addr;
+					htx->head_addr = htx->end_addr = 0;
+				}
+				else
+					htx->end_addr += sz;
+			}
+			ret = 2;
+		}
+	}
+	else if (blk->addr+sz == htx->head_addr) {
+		/* The block's payload is just before the head room */
+		if (delta < headroom) {
+			/* Expand the block's payload */
+			htx->head_addr += delta;
+			ret = 1;
+		}
+	}
+	else {
+		/* The block's payload is not at the rooms edge */
+		if (!htx->head_addr && sz+delta < tailroom) {
+			/* Move the block's payload into the tailroom */
+			if (blk->addr == htx->end_addr)
+				htx->end_addr += sz;
+			blk->addr = htx->tail_addr;
+			htx->tail_addr += sz + delta;
+			ret = 2;
+		}
+		else if (sz+delta < headroom) {
+			/* Move the block's payload into the headroom */
+			if (blk->addr == htx->end_addr)
+				htx->end_addr += sz;
+			blk->addr = htx->head_addr;
+			htx->head_addr += sz + delta;
+			ret = 2;
+		}
+	}
+	/* Otherwise defrag the HTX message */
+
+	BUG_ON((int32_t)htx->tail_addr < 0);
+	BUG_ON((int32_t)htx->head_addr < 0);
+	BUG_ON(htx->end_addr > htx->tail_addr);
+	BUG_ON(htx->head_addr > htx->end_addr);
+	return ret;
+}
+
 /* Adds a new block of type <type> in the HTTP message <htx>. Its content size
  * is passed but it is the caller responsibility to do the copy.
  */
@@ -204,6 +309,7 @@ struct htx_blk *htx_add_blk(struct htx *htx, enum htx_blk_type type, uint32_t bl
 	blk = htx_reserve_nxblk(htx, blksz);
 	if (!blk)
 		return NULL;
+	BUG_ON(blk->addr > htx->size);
 
 	blk->info = (type << 28);
 	return blk;
@@ -215,75 +321,73 @@ struct htx_blk *htx_add_blk(struct htx *htx, enum htx_blk_type type, uint32_t bl
  */
 struct htx_blk *htx_remove_blk(struct htx *htx, struct htx_blk *blk)
 {
-	enum htx_blk_type type = htx_get_blk_type(blk);
-	uint32_t next, head, pos;
+	enum htx_blk_type type;
+	uint32_t pos, addr, sz;
 
+	/* This is the last block in use */
+	if (htx->used == 1) {
+		htx_reset(htx);
+		return NULL;
+	}
+
+	type = htx_get_blk_type(blk);
+	pos  = htx_get_blk_pos(htx, blk);
+	sz   = htx_get_blksz(blk);
+	addr = blk->addr;
 	if (type != HTX_BLK_UNUSED) {
 		/* Mark the block as unused, decrement allocated size */
 		htx->data -= htx_get_blksz(blk);
 		blk->info = ((uint32_t)HTX_BLK_UNUSED << 28);
-		if (htx->sl_off == blk->addr)
-			htx->sl_off = -1;
-	}
-
-	/* This is the last block in use */
-	if (htx->used == 1/* || !htx->data */) {
-		htx->front = htx->tail = 0;
-		htx->wrap  = htx->used = 0;
-		htx->data = 0;
-		return NULL;
 	}
 
 	/* There is at least 2 blocks, so tail is always >= 0 */
-	pos  = htx_get_blk_pos(htx, blk);
-	head = htx_get_head(htx);
-	blk  = NULL;
-	next = pos + 1; /* By default retrun the next block */
-	if (htx->tail + 1 == htx->wrap) {
-		/* The HTTP message doesn't wrap */
-		if (pos == head) {
-			/* remove the head, so just return the new head */
-			htx->used--;
-			next = htx_get_head(htx);
-		}
-		else if (pos == htx->tail) {
-			/* remove the tail. this was the last inserted block so
-			 * return NULL. */
-			htx->wrap--;
-			htx->tail--;
-			htx->used--;
-			goto end;
-		}
+	if (pos == htx->head) {
+		/* move the head forward */
+		htx->used--;
+		htx->head++;
+	}
+	else if (pos == htx->tail) {
+		/* remove the tail. this was the last inserted block so
+		 * return NULL. */
+		htx->tail--;
+		htx->used--;
+		blk = NULL;
+		goto end;
+	}
+	blk = htx_get_blk(htx, pos+1);
+
+  end:
+	if (pos == htx->first)
+		htx->first = (blk ? htx_get_blk_pos(htx, blk) : -1);
+
+	if (htx->used == 1) {
+		/* If there is just one block in the HTX message, free space can
+		 * be ajusted. This operation could save some defrags. */
+		struct htx_blk *lastblk = htx_get_blk(htx, htx->tail);
+
+		htx->head_addr = 0;
+		htx->end_addr = lastblk->addr;
+		htx->tail_addr = lastblk->addr+htx->data;
 	}
 	else {
-		/* The HTTP message wraps */
-		if (pos == htx->tail) {
-			/* remove the tail. try to unwrap the message (pos == 0)
-			 * and return NULL. */
-			htx->tail = ((pos == 0) ? htx->wrap-1 : htx->tail-1);
-			htx->used--;
-			goto end;
-		}
-		else if (pos == head) {
-			/* remove the head, try to unwrap the message (pos+1 ==
-			 * wrap) and return the new head */
-			htx->used--;
-			if (pos + 1 == htx->wrap)
-				htx->wrap = htx->tail + 1;
-			next = htx_get_head(htx);
+		if (addr+sz == htx->tail_addr)
+			htx->tail_addr = addr;
+		else if (addr+sz == htx->head_addr)
+			htx->head_addr = addr;
+		if (addr == htx->end_addr) {
+			if (htx->tail_addr == htx->end_addr) {
+				htx->tail_addr = htx->head_addr;
+				htx->head_addr = htx->end_addr = 0;
+			}
+			else
+				htx->end_addr += sz;
 		}
 	}
 
-	blk = htx_get_blk(htx, next);
-	if (htx->sl_off == -1) {
-		/* Try to update the start-line offset, if possible */
-		type = htx_get_blk_type(blk);
-		if (type == HTX_BLK_REQ_SL || type == HTX_BLK_RES_SL)
-			htx->sl_off = blk->addr;
-	}
-  end:
-	if (pos == htx->front)
-		htx->front = htx_find_front(htx);
+	BUG_ON((int32_t)htx->tail_addr < 0);
+	BUG_ON((int32_t)htx->head_addr < 0);
+	BUG_ON(htx->end_addr > htx->tail_addr);
+	BUG_ON(htx->head_addr > htx->end_addr);
 	return blk;
 }
 
@@ -293,15 +397,18 @@ struct htx_blk *htx_remove_blk(struct htx *htx, struct htx_blk *blk)
 void htx_truncate(struct htx *htx, uint32_t offset)
 {
 	struct htx_blk *blk;
-	struct htx_ret htxret;
 
-	htxret = htx_find_blk(htx, offset);
-	blk = htxret.blk;
-	if (blk && htxret.ret) {
+	for (blk = htx_get_head_blk(htx); blk && offset; blk = htx_get_next_blk(htx, blk)) {
 		uint32_t sz = htx_get_blksz(blk);
+		enum htx_blk_type type = htx_get_blk_type(blk);
 
-		htx_set_blk_value_len(blk, sz - htxret.ret);
-		blk = htx_get_next_blk(htx, blk);
+		if (offset >= sz) {
+			offset -= sz;
+			continue;
+		}
+		if (type == HTX_BLK_DATA)
+			htx_change_blk_value_len(htx, blk, offset);
+		offset = 0;
 	}
 	while (blk)
 		blk = htx_remove_blk(htx, blk);
@@ -318,6 +425,12 @@ struct htx_ret htx_drain(struct htx *htx, uint32_t count)
 	struct htx_blk *blk;
 	struct htx_ret htxret = { .blk = NULL, .ret = 0 };
 
+	if (count == htx->data) {
+		htx_reset(htx);
+		htxret.ret = count;
+		return htxret;
+	}
+
 	blk = htx_get_head_blk(htx);
 	while (count && blk) {
 		uint32_t sz = htx_get_blksz(blk);
@@ -345,16 +458,17 @@ struct htx_ret htx_drain(struct htx *htx, uint32_t count)
 }
 
 /* Tries to append data to the last inserted block, if the type matches and if
- * there is enough non-wrapping space. Only DATA and TRAILERS content can be
- * appended. If the append fails, a new block is inserted. If an error occurred,
- * NULL is returned. Otherwise, on success, the updated block (or the new one)
- * is returned.
-*/
-static struct htx_blk *htx_append_blk_value(struct htx *htx, enum htx_blk_type type,
-					    const struct ist data)
+ * there is enough space to take it all. If the space wraps, the buffer is
+ * defragmented and a new block is inserted. If an error occurred, NULL is
+ * returned. Otherwise, on success, the updated block (or the new one) is
+ * returned. Due to its nature this function can be expensive and should be
+ * avoided whenever possible.
+ */
+struct htx_blk *htx_add_data_atonce(struct htx *htx, struct ist data)
 {
-	struct htx_blk *blk;
-	struct ist     v;
+	struct htx_blk *blk, *tailblk;
+	void *ptr;
+	uint32_t len, sz, tailroom, headroom;
 
 	if (!htx->used)
 		goto add_new_block;
@@ -363,60 +477,67 @@ static struct htx_blk *htx_append_blk_value(struct htx *htx, enum htx_blk_type t
 	if (data.len > htx_free_data_space(htx))
 		return NULL;
 
-	/* Append only DATA et TRAILERS data */
-	if (type != HTX_BLK_DATA && type != HTX_BLK_TLR)
+	/* get the tail block and its size */
+	tailblk = htx_get_tail_blk(htx);
+	if (tailblk == NULL)
 		goto add_new_block;
-
-	/* get the tail block */
-	blk = htx_get_blk(htx, htx->tail);
+	sz = htx_get_blksz(tailblk);
 
 	/* Don't try to append data if the last inserted block is not of the
 	 * same type */
-	if (type != htx_get_blk_type(blk))
+	if (htx_get_blk_type(tailblk) != HTX_BLK_DATA)
 		goto add_new_block;
 
 	/*
 	 * Same type and enough space: append data
 	 */
-	if (htx->tail + 1 == htx->wrap) {
-		struct htx_blk *frtblk = htx_get_blk(htx, htx->front);
-		int32_t tailroom = sizeof(htx->blocks[0]) * htx_pos_to_idx(htx, htx->tail) - (frtblk->addr + htx_get_blksz(frtblk));
-		if (tailroom >= (int32_t)data.len)
+	headroom = (htx->end_addr - htx->head_addr);
+	tailroom = sizeof(htx->blocks[0]) * htx_pos_to_idx(htx, htx->tail) - htx->tail_addr;
+	BUG_ON((int32_t)headroom < 0);
+	BUG_ON((int32_t)tailroom < 0);
+
+	len = data.len;
+	if (tailblk->addr+sz == htx->tail_addr) {
+		if (data.len <= tailroom)
 			goto append_data;
-		htx_defrag(htx, NULL);
-		blk = htx_get_blk(htx, htx->tail);
-	}
-	else {
-		struct htx_blk *headblk = htx_get_blk(htx, htx_get_head(htx));
-		int32_t headroom = headblk->addr - (blk->addr + htx_get_blksz(blk));
-		if (headroom >= (int32_t)data.len)
+		else if (!htx->head_addr) {
+			len = tailroom;
 			goto append_data;
-		htx_defrag(htx, NULL);
-		blk = htx_get_blk(htx, htx->tail);
+		}
 	}
+	else if (tailblk->addr+sz == htx->head_addr && data.len <= headroom)
+		goto append_data;
+
+	goto add_new_block;
 
   append_data:
-	/* get the value of the tail block */
 	/* FIXME: check v.len + data.len < 256MB */
-	v = htx_get_blk_value(htx, blk);
-
 	/* Append data and update the block itself */
-	memcpy(v.ptr + v.len, data.ptr, data.len);
-	htx_set_blk_value_len(blk, v.len + data.len);
+	ptr = htx_get_blk_ptr(htx, tailblk);
+	memcpy(ptr+sz, data.ptr, len);
+	htx_change_blk_value_len(htx, tailblk, sz+len);
 
-	/* Update HTTP message */
-	htx->data += data.len;
-
-	return blk;
+	if (data.len == len) {
+		blk = tailblk;
+		goto end;
+	}
+	data.ptr += len;
+	data.len -= len;
 
   add_new_block:
-	/* FIXME: check tlr.len (< 256MB) */
-	blk = htx_add_blk(htx, type, data.len);
+	/* FIXME: check data.len (< 256MB) */
+	blk = htx_add_blk(htx, HTX_BLK_DATA, data.len);
 	if (!blk)
 		return NULL;
 
 	blk->info += data.len;
 	memcpy(htx_get_blk_ptr(htx, blk), data.ptr, data.len);
+
+  end:
+	BUG_ON((int32_t)htx->tail_addr < 0);
+	BUG_ON((int32_t)htx->head_addr < 0);
+	BUG_ON(htx->end_addr > htx->tail_addr);
+	BUG_ON(htx->head_addr > htx->end_addr);
 	return blk;
 }
 
@@ -428,66 +549,79 @@ static struct htx_blk *htx_append_blk_value(struct htx *htx, enum htx_blk_type t
 struct htx_blk *htx_replace_blk_value(struct htx *htx, struct htx_blk *blk,
 				      const struct ist old, const struct ist new)
 {
-	struct buffer *tmp;
 	struct ist n, v;
 	int32_t delta;
+	int ret;
 
-	n = htx_get_blk_name(htx, blk);
-	v = htx_get_blk_value(htx, blk);
-
+	n  = htx_get_blk_name(htx, blk);
+	v  = htx_get_blk_value(htx, blk);
 	delta = new.len - old.len;
-
-	/* easy case, new data are smaller, so replace it in-place */
-	if (delta <= 0) {
-		memcpy(old.ptr, new.ptr, new.len);
-		if (old.len != v.len)
-			memmove(old.ptr + new.len, old.ptr + old.len, (v.ptr + v.len) - (old.ptr + old.len));
-		htx_set_blk_value_len(blk, v.len + delta);
-		htx->data += delta;
-		return blk;
-	}
-
-	/* we need to allocate more space to store the new header value */
-	if (delta > htx_free_space(htx))
+	ret = htx_prepare_blk_expansion(htx, blk, delta);
+	if (!ret)
 		return NULL; /* not enough space */
 
-	/*
-	 * Copy the new header in a temp buffer
-	 */
-	tmp = get_trash_chunk();
-
-	/*     1. copy the header name */
-	chunk_memcat(tmp, n.ptr, n.len);
-
-	/*     2. copy value before old part, if any */
-	if (old.ptr != v.ptr)
-		chunk_memcat(tmp, v.ptr, old.ptr - v.ptr);
-
-	/*     3. copy new value */
-	chunk_memcat(tmp, new.ptr, new.len);
-
-	/*     4. copy value after old part if any */
-	if (old.len != v.len)
-		chunk_memcat(tmp, old.ptr + old.len, (v.ptr + v.len) - (old.ptr + old.len));
-
-
-	/* Expand the block size. But data are not copied yet. Then defragment
-	 * the HTX message.
-	 */
+	/* Before updating the payload, set the new block size and update HTX
+	 * message */
 	htx_set_blk_value_len(blk, v.len + delta);
 	htx->data += delta;
-	blk = htx_defrag(htx, blk);
 
-	/* Finaly, copy data. */
-	memcpy(htx_get_blk_ptr(htx, blk), tmp->area, tmp->data);
+	if (ret == 1) {
+		if (delta <= 0) {
+			/* compression: copy new data first */
+			memcpy(old.ptr, new.ptr, new.len);
+			memmove(old.ptr + new.len, old.ptr + old.len, (v.ptr + v.len) - (old.ptr + old.len));
+		}
+		else {
+			/* expansion: move the end first */
+			memmove(old.ptr + new.len, old.ptr + old.len, (v.ptr + v.len) - (old.ptr + old.len));
+			memcpy(old.ptr, new.ptr, new.len);
+		}
+	}
+	else if (ret == 2) {
+		void *ptr = htx_get_blk_ptr(htx, blk);
 
+		/* Copy the name, if any */
+		memcpy(ptr, n.ptr, n.len);
+		ptr += n.len;
+
+		/* Copy value before old part, if any */
+		memcpy(ptr, v.ptr, old.ptr - v.ptr);
+		ptr += old.ptr - v.ptr;
+
+		/* Copy new value */
+		memcpy(ptr, new.ptr, new.len);
+		ptr += new.len;
+
+		/* Copy value after old part, if any */
+		memcpy(ptr, old.ptr + old.len, (v.ptr + v.len) - (old.ptr + old.len));
+	}
+	else {
+		struct buffer *tmp = get_trash_chunk();
+
+		/* Copy the header name, if any */
+		chunk_memcat(tmp, n.ptr, n.len);
+
+		/* Copy value before old part, if any */
+		chunk_memcat(tmp, v.ptr, old.ptr - v.ptr);
+
+		/* Copy new value */
+		chunk_memcat(tmp, new.ptr, new.len);
+
+		/* Copy value after old part if any */
+                chunk_memcat(tmp, old.ptr + old.len, (v.ptr + v.len) - (old.ptr + old.len));
+
+		blk = htx_defrag(htx, blk);
+
+		/* Finaly, copy data. */
+		memcpy(htx_get_blk_ptr(htx, blk), tmp->area, tmp->data);
+	}
 	return blk;
 }
 
 /* Transfer HTX blocks from <src> to <dst>, stopping on the first block of the
- * type <mark> (typically EOH, EOD or EOM) or when <count> bytes of data were
- * moved. It returns the number of bytes of data moved and the last HTX block
- * inserted in <dst>.
+ * type <mark> (typically EOH or EOM) or when <count> bytes were moved
+ * (including payload and meta-data). It returns the number of bytes moved and
+ * the last HTX block inserted in <dst>.
  */
 struct htx_ret htx_xfer_blks(struct htx *dst, struct htx *src, uint32_t count,
 			     enum htx_blk_type mark)
@@ -495,36 +629,53 @@ struct htx_ret htx_xfer_blks(struct htx *dst, struct htx *src, uint32_t count,
 	struct htx_blk   *blk, *dstblk;
 	enum htx_blk_type type;
 	uint32_t	  info, max, sz, ret;
+	int inside_trailers = 0;
 
-	ret = 0;
+	ret = htx_used_space(dst);
 	blk = htx_get_blk(src, htx_get_head(src));
 	dstblk = NULL;
-	while (blk && ret <= count) {
+
+	while (blk && count) {
 		type = htx_get_blk_type(blk);
 
 		/* Ingore unused block */
 		if (type == HTX_BLK_UNUSED)
 			goto next;
 
-		sz = htx_get_blksz(blk);
-		if (!sz) {
-			dstblk = htx_reserve_nxblk(dst, 0);
-			if (!dstblk)
+		/* Be sure to have enough space to xfer all headers/trailers in
+		 * one time. If not while <dst> is empty, we report a parsing
+		 * error on <src>.
+		 */
+		if (mark >= HTX_BLK_EOH && (type == HTX_BLK_REQ_SL || type == HTX_BLK_RES_SL)) {
+			struct htx_sl *sl = htx_get_blk_ptr(src, blk);
+
+			if (sl->hdrs_bytes != -1 && sl->hdrs_bytes > count) {
+				if (htx_is_empty(dst))
+					src->flags |= HTX_FL_PARSING_ERROR;
 				break;
-			dstblk->info = blk->info;
-			goto next;
+			}
+		}
+		else if ((type == HTX_BLK_TLR || type == HTX_BLK_EOT) &&
+			 !inside_trailers && mark >= HTX_BLK_EOT) {
+			inside_trailers = 1;
+			if (htx_used_space(src) > count) {
+				if (htx_is_empty(dst))
+					src->flags |= HTX_FL_PARSING_ERROR;
+				break;
+			}
 		}
 
+		sz = htx_get_blksz(blk);
 		info = blk->info;
-		max = htx_free_data_space(dst);
-		if (max > count)
-			max = count;
-		if (ret + sz > max) {
-			sz = max - ret;
-			info = (type << 28) + sz;
-			/* Headers and pseudo headers must be fully copied  */
-			if (type < HTX_BLK_DATA || !sz)
+		max = htx_get_max_blksz(dst, count);
+		if (!max)
+			break;
+		if (sz > max) {
+			/* Headers must be fully copied  */
+			if (type != HTX_BLK_DATA)
 				break;
+			sz = max;
+			info = (type << 28) + sz;
 		}
 
 		dstblk = htx_reserve_nxblk(dst, sz);
@@ -533,18 +684,13 @@ struct htx_ret htx_xfer_blks(struct htx *dst, struct htx *src, uint32_t count,
 		dstblk->info = info;
 		memcpy(htx_get_blk_ptr(dst, dstblk), htx_get_blk_ptr(src, blk), sz);
 
-		ret += sz;
+		count -= sizeof(dstblk) + sz;
 		if (blk->info != info) {
 			/* Partial move: don't remove <blk> from <src> but
 			 * resize its content */
-			blk->addr += sz;
-			htx_set_blk_value_len(blk, htx_get_blksz(blk) - sz);
-			src->data -= sz;
+			htx_cut_data_blk(src, blk, sz);
 			break;
 		}
-
-		if (dst->sl_off == -1 && src->sl_off == blk->addr)
-			dst->sl_off = dstblk->addr;
 	  next:
 		blk = htx_remove_blk(src, blk);
 		if (type == mark)
@@ -552,6 +698,8 @@ struct htx_ret htx_xfer_blks(struct htx *dst, struct htx *src, uint32_t count,
 
 	}
 
+  end:
+	ret = htx_used_space(dst) - ret;
 	return (struct htx_ret){.ret = ret, .blk = dstblk};
 }
 
@@ -563,36 +711,30 @@ struct htx_blk *htx_replace_header(struct htx *htx, struct htx_blk *blk,
 				   const struct ist name, const struct ist value)
 {
 	enum htx_blk_type type;
+	void *ptr;
 	int32_t delta;
+	int ret;
 
 	type = htx_get_blk_type(blk);
 	if (type != HTX_BLK_HDR)
 		return NULL;
 
 	delta = name.len + value.len - htx_get_blksz(blk);
-
-	/* easy case, new value is smaller, so replace it in-place */
-	if (delta <= 0) {
-		blk->info = (type << 28) + (value.len << 8) + name.len;
-		htx->data += delta;
-		goto copy;
-	}
-
-	/* we need to allocate more space to store the new value */
-	if (delta > htx_free_space(htx))
+	ret = htx_prepare_blk_expansion(htx, blk, delta);
+	if (!ret)
 		return NULL; /* not enough space */
 
-	/* Expand the block size. But data are not copied yet. Then defragment
-	 * the HTX message.
-	 */
+	/* Set the new block size and update HTX message */
 	blk->info = (type << 28) + (value.len << 8) + name.len;
 	htx->data += delta;
-	blk = htx_defrag(htx, blk);
 
-  copy:
+	if (ret == 3)
+		blk = htx_defrag(htx, blk);
+
 	/* Finaly, copy data. */
-	ist2bin_lc(htx_get_blk_ptr(htx, blk), name);
-	memcpy(htx_get_blk_ptr(htx, blk) + name.len, value.ptr, value.len);
+	ptr = htx_get_blk_ptr(htx, blk);
+	ist2bin_lc(ptr, name);
+	memcpy(ptr + name.len, value.ptr, value.len);
 	return blk;
 }
 
@@ -603,11 +745,12 @@ struct htx_blk *htx_replace_header(struct htx *htx, struct htx_blk *blk,
 struct htx_sl *htx_replace_stline(struct htx *htx, struct htx_blk *blk, const struct ist p1,
 				  const struct ist p2, const struct ist p3)
 {
+	enum htx_blk_type type;
 	struct htx_sl *sl;
 	struct htx_sl tmp; /* used to save sl->info and sl->flags */
-	enum htx_blk_type type;
-	uint32_t size;
+	uint32_t sz;
 	int32_t delta;
+	int ret;
 
 	type = htx_get_blk_type(blk);
 	if (type != HTX_BLK_REQ_SL && type != HTX_BLK_RES_SL)
@@ -617,38 +760,26 @@ struct htx_sl *htx_replace_stline(struct htx *htx, struct htx_blk *blk, const st
 	sl = htx_get_blk_ptr(htx, blk);
 	tmp.info = sl->info;
 	tmp.flags = sl->flags;
-	if (htx->sl_off == blk->addr)
-		htx->sl_off = -1;
+	tmp.hdrs_bytes = sl->hdrs_bytes;
 
-
-	size = sizeof(*sl) + p1.len + p2.len + p3.len;
-	delta = size - htx_get_blksz(blk);
-
-	/* easy case, new data are smaller, so replace it in-place */
-	if (delta <= 0) {
-		htx_set_blk_value_len(blk, size);
-		htx->data += delta;
-		goto copy;
-	}
-
-	/* we need to allocate more space to store the new value */
-	if (delta > htx_free_space(htx))
+	sz = htx_get_blksz(blk);
+	delta = sizeof(*sl) + p1.len + p2.len + p3.len - sz;
+	ret = htx_prepare_blk_expansion(htx, blk, delta);
+	if (!ret)
 		return NULL; /* not enough space */
 
-	/* Expand the block size. But data are not copied yet. Then defragment
-	 * the HTX message.
-	 */
-	htx_set_blk_value_len(blk, size);
+	/* Set the new block size and update HTX message */
+	htx_set_blk_value_len(blk, sz+delta);
 	htx->data += delta;
-	blk = htx_defrag(htx, blk);
 
-  copy:
+	if (ret == 3)
+		blk = htx_defrag(htx, blk);
+
 	/* Restore start-line info and flags and copy parts of the start-line */
 	sl = htx_get_blk_ptr(htx, blk);
 	sl->info = tmp.info;
 	sl->flags = tmp.flags;
-	if (htx->sl_off == -1)
-		htx->sl_off = blk->addr;
+	sl->hdrs_bytes = tmp.hdrs_bytes;
 
 	HTX_SL_P1_LEN(sl) = p1.len;
 	HTX_SL_P2_LEN(sl) = p2.len;
@@ -683,9 +814,7 @@ struct htx_sl *htx_add_stline(struct htx *htx, enum htx_blk_type type, unsigned
 	blk->info += size;
 
 	sl = htx_get_blk_ptr(htx, blk);
-	if (htx->sl_off == -1)
-		htx->sl_off = blk->addr;
-
+	sl->hdrs_bytes = -1;
 	sl->flags = flags;
 
 	HTX_SL_P1_LEN(sl) = p1.len;
@@ -718,6 +847,25 @@ struct htx_blk *htx_add_header(struct htx *htx, const struct ist name,
 	return blk;
 }
 
+/* Adds an HTX block of type TLR in <htx>. It returns the new block on
+ * success. Otherwise, it returns NULL. The header name is always lower cased.
+ */
+struct htx_blk *htx_add_trailer(struct htx *htx, const struct ist name,
+				const struct ist value)
+{
+	struct htx_blk *blk;
+
+	/* FIXME: check name.len (< 256B) and value.len (< 1MB) */
+	blk = htx_add_blk(htx, HTX_BLK_TLR, name.len + value.len);
+	if (!blk)
+		return NULL;
+
+	blk->info += (value.len << 8) + name.len;
+	ist2bin_lc(htx_get_blk_ptr(htx, blk), name);
+	memcpy(htx_get_blk_ptr(htx, blk)  + name.len, value.ptr, value.len);
+	return blk;
+}
+
 /* Adds an HTX block of type <type> in <htx>, of size <blksz>. It returns the
  * new block on success. Otherwise, it returns NULL. The caller is responsible
  * for filling the block itself.
@@ -744,25 +892,19 @@ struct htx_blk *htx_add_all_headers(struct htx *htx, const struct http_hdr *hdrs
 	}
 	return htx_add_endof(htx, HTX_BLK_EOH);
 }
-/* Adds an HTX block of type PHDR in <htx>. It returns the new block on
- * success. Otherwise, it returns NULL.
- */
-struct htx_blk *htx_add_pseudo_header(struct htx *htx,  enum htx_phdr_type phdr,
-				      const struct ist value)
+
+struct htx_blk *htx_add_all_trailers(struct htx *htx, const struct http_hdr *hdrs)
 {
-	struct htx_blk *blk;
+	int i;
 
-	/* FIXME: check value.len ( < 1MB) */
-	blk = htx_add_blk(htx, HTX_BLK_PHDR, value.len);
-	if (!blk)
-		return NULL;
-
-	blk->info += (value.len << 8) + phdr;
-	memcpy(htx_get_blk_ptr(htx, blk), value.ptr, value.len);
-	return blk;
+	for (i = 0; hdrs[i].n.len; i++) {
+		if (!htx_add_trailer(htx, hdrs[i].n, hdrs[i].v))
+			return NULL;
+	}
+	return htx_add_endof(htx, HTX_BLK_EOT);
 }
 
-/* Adds an HTX block of type EOH,EOD or EOM in <htx>. It returns the new block
+/* Adds an HTX block of type EOH or EOM in <htx>. It returns the new block
  * on success. Otherwise, it returns NULL.
  */
 struct htx_blk *htx_add_endof(struct htx *htx, enum htx_blk_type type)
@@ -779,55 +921,96 @@ struct htx_blk *htx_add_endof(struct htx *htx, enum htx_blk_type type)
 
 
 /* Adds an HTX block of type DATA in <htx>. It first tries to append data if
- * possible. It returns the new block on success. Otherwise, it returns NULL.
+ * possible. It returns the number of bytes consumed from <data>, which may be
+ * zero if nothing could be copied.
  */
-struct htx_blk *htx_add_data(struct htx *htx, const struct ist data)
+size_t htx_add_data(struct htx *htx, const struct ist data)
 {
-	return htx_append_blk_value(htx, HTX_BLK_DATA, data);
-}
+	struct htx_blk *blk, *tailblk;
+	void *ptr;
+	uint32_t sz, room;
+	int32_t len = data.len;
 
-/* Adds an HTX block of type TLR in <htx>. It first tries to append trailers
- * data if possible. It returns the new block on success. Otherwise, it returns
- * NULL.
- */
-struct htx_blk *htx_add_trailer(struct htx *htx, const struct ist tlr)
-{
-	return htx_append_blk_value(htx, HTX_BLK_TLR, tlr);
-}
+	if (!htx->used)
+		goto add_new_block;
 
-/* Adds an HTX block of type OOB in <htx>. It returns the new block on
- * success. Otherwise, it returns NULL.
- */
-struct htx_blk *htx_add_oob(struct htx *htx, const struct ist oob)
-{
-	struct htx_blk *blk;
+	/* Not enough space to store data */
+	if (len > htx_free_data_space(htx))
+		len = htx_free_data_space(htx);
 
-	/* FIXME: check oob.len (< 256MB) */
-	blk = htx_add_blk(htx, HTX_BLK_OOB, oob.len);
-	if (!blk)
-		return NULL;
+	if (!len)
+		return 0;
 
-	blk->info += oob.len;
-	memcpy(htx_get_blk_ptr(htx, blk), oob.ptr, oob.len);
-	return blk;
-}
+	/* get the tail and head block */
+	tailblk = htx_get_tail_blk(htx);
+	if (tailblk == NULL)
+		goto add_new_block;
+	sz = htx_get_blksz(tailblk);
 
-struct htx_blk *htx_add_data_before(struct htx *htx, const struct htx_blk *ref,
-				    const struct ist data)
-{
-	struct htx_blk *blk;
-	int32_t prev;
+	/* Don't try to append data if the last inserted block is not of the
+	 * same type */
+	if (htx_get_blk_type(tailblk) != HTX_BLK_DATA)
+		goto add_new_block;
 
+	/*
+	 * Same type and enough space: append data
+	 */
+	if (!htx->head_addr) {
+		if (tailblk->addr+sz != htx->tail_addr)
+			goto add_new_block;
+		room = sizeof(htx->blocks[0]) * htx_pos_to_idx(htx, htx->tail) - htx->tail_addr;
+	}
+	else {
+		if (tailblk->addr+sz != htx->head_addr)
+			goto add_new_block;
+		room = (htx->end_addr - htx->head_addr);
+	}
+	BUG_ON((int32_t)room < 0);
+	if (room < len)
+		len = room;
+
+  append_data:
+	/* FIXME: check v.len + data.len < 256MB */
+	/* Append data and update the block itself */
+	ptr = htx_get_blk_ptr(htx, tailblk);
+	memcpy(ptr + sz, data.ptr, len);
+	htx_change_blk_value_len(htx, tailblk, sz+len);
+
+	BUG_ON((int32_t)htx->tail_addr < 0);
+	BUG_ON((int32_t)htx->head_addr < 0);
+	BUG_ON(htx->end_addr > htx->tail_addr);
+	BUG_ON(htx->head_addr > htx->end_addr);
+	return len;
+
+  add_new_block:
 	/* FIXME: check data.len (< 256MB) */
-	blk = htx_add_blk(htx, HTX_BLK_DATA, data.len);
+	blk = htx_add_blk(htx, HTX_BLK_DATA, len);
+	if (!blk)
+		return 0;
+
+	blk->info += len;
+	memcpy(htx_get_blk_ptr(htx, blk), data.ptr, len);
+	return len;
+}
+
+
+/* Adds an HTX block of type DATA in <htx> just after all other DATA
+ * blocks. Because it relies on htx_add_data_atonce(), It may be happened to a
+ * DATA block if possible. But, if the function succeeds, it will be the last
+ * DATA block in all cases. If an error occurred, NULL is returned. Otherwise,
+ * on success, the updated block (or the new one) is returned.
+ */
+struct htx_blk *htx_add_last_data(struct htx *htx, struct ist data)
+{
+	struct htx_blk *blk, *pblk;
+
+	blk = htx_add_data_atonce(htx, data);
 	if (!blk)
 		return NULL;
 
-	blk->info += data.len;
-	memcpy(htx_get_blk_ptr(htx, blk), data.ptr, data.len);
-
-	for (prev = htx_get_prev(htx, htx->tail); prev != -1; prev = htx_get_prev(htx, prev)) {
-		struct htx_blk *pblk = htx_get_blk(htx, prev);
+	for (pblk = htx_get_prev_blk(htx, blk); pblk; pblk = htx_get_prev_blk(htx, pblk)) {
+		if (htx_get_blk_type(pblk) <= HTX_BLK_DATA)
+			break;
 
 		/* Swap .addr and .info fields */
 		blk->addr ^= pblk->addr; pblk->addr ^= blk->addr; blk->addr ^= pblk->addr;
@@ -835,15 +1018,35 @@ struct htx_blk *htx_add_data_before(struct htx *htx, const struct htx_blk *ref,
 
 		if (blk->addr == pblk->addr)
 			blk->addr += htx_get_blksz(pblk);
-		htx->front = prev;
-
-		if (pblk == ref)
-			break;
 		blk = pblk;
 	}
+
 	return blk;
 }
 
+/* Moves the block <blk> just before the block <ref>. Both blocks must be in the
+ * HTX message <htx> and <blk> must be placed after <ref>. pointer to these
+ * blocks are updated to remain valid after the move. */
+void htx_move_blk_before(struct htx *htx, struct htx_blk **blk, struct htx_blk **ref)
+{
+	struct htx_blk *cblk, *pblk;
+
+	cblk = *blk;
+	for (pblk = htx_get_prev_blk(htx, cblk); pblk; pblk = htx_get_prev_blk(htx, pblk)) {
+		/* Swap .addr and .info fields */
+		cblk->addr ^= pblk->addr; pblk->addr ^= cblk->addr; cblk->addr ^= pblk->addr;
+		cblk->info ^= pblk->info; pblk->info ^= cblk->info; cblk->info ^= pblk->info;
+
+		if (cblk->addr == pblk->addr)
+			cblk->addr += htx_get_blksz(pblk);
+		if (pblk == *ref)
+			break;
+		cblk = pblk;
+	}
+	*blk = cblk;
+	*ref = pblk;
+}
+
 /* Appends the H1 representation of the request line block <blk> to the
  * chunk <chk>. It returns 1 if data are successfully appended, otherwise it
  * returns 0.
@@ -939,15 +1142,3 @@ int htx_data_to_h1(const struct ist data, struct buffer *chk, int chunked)
 
 	return 1;
 }
-
-/* Appends the h1 representation of the trailer block <blk> to the chunk
- * <chk>. It returns 1 if data are successfully appended, otherwise it returns
- * 0.
- */
-int htx_trailer_to_h1(const struct ist tlr, struct buffer *chk)
-{
-	/* FIXME: be sure the CRLF is here or remove it when inserted */
-	if (!chunk_memcat(chk, tlr.ptr, tlr.len))
-		return 0;
-	return 1;
-}
