@@ -30,8 +30,10 @@
 #include <proto/applet.h>
 #include <proto/channel.h>
 #include <proto/connection.h>
+#include <proto/http_htx.h>
 #include <proto/mux_pt.h>
 #include <proto/pipe.h>
+#include <proto/proxy.h>
 #include <proto/stream.h>
 #include <proto/stream_interface.h>
 #include <proto/task.h>
@@ -168,13 +170,12 @@ static void stream_int_shutr(struct stream_interface *si)
 	struct channel *ic = si_ic(si);
 
 	si_rx_shut_blk(si);
-	ic->flags &= ~CF_SHUTR_NOW;
 	if (ic->flags & CF_SHUTR)
 		return;
 	ic->flags |= CF_SHUTR;
 	ic->rex = TICK_ETERNITY;
 
-	if (si->state != SI_ST_EST && si->state != SI_ST_CON)
+	if (!si_state_in(si->state, SI_SB_CON|SI_SB_RDY|SI_SB_EST))
 		return;
 
 	if (si_oc(si)->flags & CF_SHUTW) {
@@ -216,6 +217,7 @@ static void stream_int_shutw(struct stream_interface *si)
 	}
 
 	switch (si->state) {
+	case SI_ST_RDY:
 	case SI_ST_EST:
 		/* we have to shut before closing, otherwise some short messages
 		 * may never leave the system, especially when there are remaining
@@ -237,7 +239,6 @@ static void stream_int_shutw(struct stream_interface *si)
 	default:
 		si->flags &= ~SI_FL_NOLINGER;
 		si_rx_shut_blk(si);
-		ic->flags &= ~CF_SHUTR_NOW;
 		ic->flags |= CF_SHUTR;
 		ic->rex = TICK_ETERNITY;
 		si->exp = TICK_ETERNITY;
@@ -263,7 +264,7 @@ static void stream_int_chk_rcv(struct stream_interface *si)
 	}
 	else {
 		/* (re)start reading */
-		tasklet_wakeup(si->wait_event.task);
+		tasklet_wakeup(si->wait_event.tasklet);
 		if (!(si->flags & SI_FL_DONT_WAKE))
 			task_wakeup(si_task(si), TASK_WOKEN_IO);
 	}
@@ -339,7 +340,7 @@ int conn_si_send_proxy(struct connection *conn, unsigned int flag)
 	 * connection, in which case the connection is validated only once
 	 * we've sent the whole proxy line. Otherwise we use connect().
 	 */
-	while (conn->send_proxy_ofs) {
+	if (conn->send_proxy_ofs) {
 		const struct conn_stream *cs;
 		int ret;
 
@@ -360,6 +361,21 @@ int conn_si_send_proxy(struct connection *conn, unsigned int flag)
 			ret = make_proxy_line(trash.area, trash.size,
 					      objt_server(conn->target),
 					      remote_cs ? remote_cs->conn : NULL);
+			/* We may not have a conn_stream yet, if we don't
+			 * know which mux to use, because it will be decided
+			 * during the SSL handshake. In this case, there should
+			 * be a session associated to the connection in
+			 * conn->owner, and we know it is the session that
+			 * initiated that connection, so we can just use
+			 * its origin, which should contain the client
+			 * connection.
+			 */
+		} else if (!cs && conn->owner) {
+			struct session *sess = conn->owner;
+
+			ret = make_proxy_line(trash.area, trash.size,
+			                      objt_server(conn->target),
+					      objt_conn(sess->origin));
 		}
 		else {
 			/* The target server expects a LOCAL line to be sent first. Retrieving
@@ -399,7 +415,6 @@ int conn_si_send_proxy(struct connection *conn, unsigned int flag)
 			goto out_wait;
 
 		/* OK we've sent the whole line, we're connected */
-		break;
 	}
 
 	/* The connection is ready now, simply return and let the connection
@@ -416,7 +431,6 @@ int conn_si_send_proxy(struct connection *conn, unsigned int flag)
 	return 0;
 
  out_wait:
-	__conn_sock_stop_recv(conn);
 	return 0;
 }
 
@@ -528,7 +542,7 @@ static void stream_int_notify(struct stream_interface *si)
 	/* wake the task up only when needed */
 	if (/* changes on the production side */
 	    (ic->flags & (CF_READ_NULL|CF_READ_ERROR)) ||
-	    (si->state != SI_ST_EST && si->state != SI_ST_CON) ||
+	    !si_state_in(si->state, SI_SB_CON|SI_SB_RDY|SI_SB_EST) ||
 	    (si->flags & SI_FL_ERR) ||
 	    ((ic->flags & CF_READ_PARTIAL) &&
 	     ((ic->flags & CF_EOI) || !ic->to_forward || sio->state != SI_ST_EST)) ||
@@ -574,8 +588,14 @@ static int si_cs_process(struct conn_stream *cs)
 
 	/* First step, report to the stream-int what was detected at the
 	 * connection layer : errors and connection establishment.
+	 * Only add SI_FL_ERR if we're connected, or we're attempting to
+	 * connect, we may get there because we got woken up, but only run
+	 * after process_stream() noticed there were an error, and decided
+	 * to retry to connect, the connection may still have CO_FL_ERROR,
+	 * and we don't want to add SI_FL_ERR back
 	 */
-	if (conn->flags & CO_FL_ERROR || cs->flags & CS_FL_ERROR)
+	if (si->state >= SI_ST_CON &&
+	    (conn->flags & CO_FL_ERROR || cs->flags & CS_FL_ERROR))
 		si->flags |= SI_FL_ERR;
 
 	/* If we had early data, and the handshake ended, then
@@ -589,16 +609,18 @@ static int si_cs_process(struct conn_stream *cs)
 		task_wakeup(si_task(si), TASK_WOKEN_MSG);
 	}
 
-	if ((si->state < SI_ST_EST) &&
+	if (!si_state_in(si->state, SI_SB_EST|SI_SB_DIS|SI_SB_CLO) &&
 	    (conn->flags & (CO_FL_CONNECTED | CO_FL_HANDSHAKE)) == CO_FL_CONNECTED) {
 		si->exp = TICK_ETERNITY;
 		oc->flags |= CF_WRITE_NULL;
+		if (si->state == SI_ST_CON)
+			si->state = SI_ST_RDY;
 	}
 
 	/* Report EOI on the channel if it was reached from the mux point of
 	 * view. */
 	if ((cs->flags & CS_FL_EOI) && !(ic->flags & CF_EOI))
-		ic->flags |= CF_EOI;
+		ic->flags |= (CF_EOI|CF_READ_PARTIAL);
 
 	/* Second step : update the stream-int and channels, try to forward any
 	 * pending data, then possibly wake the stream up based on the new
@@ -629,6 +651,14 @@ int si_cs_send(struct conn_stream *cs)
 		return 0;
 
 	if (conn->flags & CO_FL_ERROR || cs->flags & (CS_FL_ERROR|CS_FL_ERR_PENDING)) {
+		/* We're probably there because the tasklet was woken up,
+		 * but process_stream() ran before, detected there were an
+		 * error and put the si back to SI_ST_TAR. There's still
+		 * CO_FL_ERROR on the connection but we don't want to add
+		 * SI_FL_ERR back, so give up
+		 */
+		if (si->state < SI_ST_CON)
+			return 0;
 		si->flags |= SI_FL_ERR;
 		return 1;
 	}
@@ -642,6 +672,8 @@ int si_cs_send(struct conn_stream *cs)
 		if (ret > 0) {
 			oc->flags |= CF_WRITE_PARTIAL | CF_WROTE_DATA;
 			did_send = 1;
+			if (si->state == SI_ST_CON)
+				si->state = SI_ST_RDY;
 		}
 
 		if (!oc->pipe->data) {
@@ -688,10 +720,41 @@ int si_cs_send(struct conn_stream *cs)
 		if (oc->flags & CF_STREAMER)
 			send_flag |= CO_SFL_STREAMER;
 
+		if ((si->flags & SI_FL_L7_RETRY) && !b_data(&si->l7_buffer)) {
+			struct stream *s = si_strm(si);
+			/* If we want to be able to do L7 retries, copy
+			 * the data we're about to send, so that we are able
+			 * to resend them if needed
+			 */
+			/* Try to allocate a buffer if we had none.
+			 * If it fails, the next test will just
+			 * disable the l7 retries by setting
+			 * l7_conn_retries to 0.
+			 */
+			if (!s->txn || (s->txn->req.msg_state != HTTP_MSG_DONE))
+				si->flags &= ~SI_FL_L7_RETRY;
+			else {
+				if (b_is_null(&si->l7_buffer))
+					b_alloc(&si->l7_buffer);
+				if (b_is_null(&si->l7_buffer))
+					si->flags &= ~SI_FL_L7_RETRY;
+				else {
+					memcpy(b_orig(&si->l7_buffer),
+					       b_orig(&oc->buf),
+					       b_size(&oc->buf));
+					si->l7_buffer.head = co_data(oc);
+					b_add(&si->l7_buffer, co_data(oc));
+				}
+
+			}
+		}
+
 		ret = cs->conn->mux->snd_buf(cs, &oc->buf, co_data(oc), send_flag);
 		if (ret > 0) {
 			did_send = 1;
 			oc->flags |= CF_WRITE_PARTIAL | CF_WROTE_DATA;
+			if (si->state == SI_ST_CON)
+				si->state = SI_ST_RDY;
 
 			co_set_data(oc, co_data(oc) - ret);
 			c_realign_if_empty(oc);
@@ -744,130 +807,154 @@ struct task *si_cs_io_cb(struct task *t, void *ctx, unsigned short state)
 }
 
 /* This function is designed to be called from within the stream handler to
- * update the channels' expiration timers and the stream interface's flags
- * based on the channels' flags. It needs to be called only once after the
- * channels' flags have settled down, and before they are cleared, though it
- * doesn't harm to call it as often as desired (it just slightly hurts
- * performance). It must not be called from outside of the stream handler,
- * as what it does will be used to compute the stream task's expiration.
+ * update the input channel's expiration timer and the stream interface's
+ * Rx flags based on the channel's flags. It needs to be called only once
+ * after the channel's flags have settled down, and before they are cleared,
+ * though it doesn't harm to call it as often as desired (it just slightly
+ * hurts performance). It must not be called from outside of the stream
+ * handler, as what it does will be used to compute the stream task's
+ * expiration.
  */
-void si_update(struct stream_interface *si)
+void si_update_rx(struct stream_interface *si)
 {
 	struct channel *ic = si_ic(si);
-	struct channel *oc = si_oc(si);
 
-	if (!(ic->flags & CF_SHUTR)) {
-		/* Read not closed, update FD status and timeout for reads */
-		if (ic->flags & CF_DONT_READ)
-			si_rx_chan_blk(si);
-		else
-			si_rx_chan_rdy(si);
-
-		if (!channel_is_empty(ic)) {
-			/* stop reading, imposed by channel's policy or contents */
-			si_rx_room_blk(si);
-		}
-		else {
-			/* (re)start reading and update timeout. Note: we don't recompute the timeout
-			 * everytime we get here, otherwise it would risk never to expire. We only
-			 * update it if is was not yet set. The stream socket handler will already
-			 * have updated it if there has been a completed I/O.
-			 */
-			si_rx_room_rdy(si);
-		}
-		if (si->flags & SI_FL_RXBLK_ANY & ~SI_FL_RX_WAIT_EP)
-			ic->rex = TICK_ETERNITY;
-		else if (!(ic->flags & CF_READ_NOEXP) && !tick_isset(ic->rex))
-			ic->rex = tick_add_ifset(now_ms, ic->rto);
-
-		si_chk_rcv(si);
-	}
-	else
+	if (ic->flags & CF_SHUTR) {
 		si_rx_shut_blk(si);
+		return;
+	}
 
-	if (!(oc->flags & CF_SHUTW)) {
-		/* Write not closed, update FD status and timeout for writes */
-		if (channel_is_empty(oc)) {
-			/* stop writing */
-			if (!(si->flags & SI_FL_WAIT_DATA)) {
-				if ((oc->flags & CF_SHUTW_NOW) == 0)
-					si->flags |= SI_FL_WAIT_DATA;
-				oc->wex = TICK_ETERNITY;
-			}
+	/* Read not closed, update FD status and timeout for reads */
+	if (ic->flags & CF_DONT_READ)
+		si_rx_chan_blk(si);
+	else
+		si_rx_chan_rdy(si);
+
+	if (!channel_is_empty(ic)) {
+		/* stop reading, imposed by channel's policy or contents */
+		si_rx_room_blk(si);
+	}
+	else {
+		/* (re)start reading and update timeout. Note: we don't recompute the timeout
+		 * everytime we get here, otherwise it would risk never to expire. We only
+		 * update it if is was not yet set. The stream socket handler will already
+		 * have updated it if there has been a completed I/O.
+		 */
+		si_rx_room_rdy(si);
+	}
+	if (si->flags & SI_FL_RXBLK_ANY & ~SI_FL_RX_WAIT_EP)
+		ic->rex = TICK_ETERNITY;
+	else if (!(ic->flags & CF_READ_NOEXP) && !tick_isset(ic->rex))
+		ic->rex = tick_add_ifset(now_ms, ic->rto);
+
+	si_chk_rcv(si);
+}
+
+/* This function is designed to be called from within the stream handler to
+ * update the output channel's expiration timer and the stream interface's
+ * Tx flags based on the channel's flags. It needs to be called only once
+ * after the channel's flags have settled down, and before they are cleared,
+ * though it doesn't harm to call it as often as desired (it just slightly
+ * hurts performance). It must not be called from outside of the stream
+ * handler, as what it does will be used to compute the stream task's
+ * expiration.
+ */
+void si_update_tx(struct stream_interface *si)
+{
+	struct channel *oc = si_oc(si);
+	struct channel *ic = si_ic(si);
+
+	if (oc->flags & CF_SHUTW)
+		return;
+
+	/* Write not closed, update FD status and timeout for writes */
+	if (channel_is_empty(oc)) {
+		/* stop writing */
+		if (!(si->flags & SI_FL_WAIT_DATA)) {
+			if ((oc->flags & CF_SHUTW_NOW) == 0)
+				si->flags |= SI_FL_WAIT_DATA;
+			oc->wex = TICK_ETERNITY;
 		}
-		else {
-			/* (re)start writing and update timeout. Note: we don't recompute the timeout
-			 * everytime we get here, otherwise it would risk never to expire. We only
-			 * update it if is was not yet set. The stream socket handler will already
-			 * have updated it if there has been a completed I/O.
+		return;
+	}
+
+	/* (re)start writing and update timeout. Note: we don't recompute the timeout
+	 * everytime we get here, otherwise it would risk never to expire. We only
+	 * update it if is was not yet set. The stream socket handler will already
+	 * have updated it if there has been a completed I/O.
+	 */
+	si->flags &= ~SI_FL_WAIT_DATA;
+	if (!tick_isset(oc->wex)) {
+		oc->wex = tick_add_ifset(now_ms, oc->wto);
+		if (tick_isset(ic->rex) && !(si->flags & SI_FL_INDEP_STR)) {
+			/* Note: depending on the protocol, we don't know if we're waiting
+			 * for incoming data or not. So in order to prevent the socket from
+			 * expiring read timeouts during writes, we refresh the read timeout,
+			 * except if it was already infinite or if we have explicitly setup
+			 * independent streams.
 			 */
-			si->flags &= ~SI_FL_WAIT_DATA;
-			if (!tick_isset(oc->wex)) {
-				oc->wex = tick_add_ifset(now_ms, oc->wto);
-				if (tick_isset(ic->rex) && !(si->flags & SI_FL_INDEP_STR)) {
-					/* Note: depending on the protocol, we don't know if we're waiting
-					 * for incoming data or not. So in order to prevent the socket from
-					 * expiring read timeouts during writes, we refresh the read timeout,
-					 * except if it was already infinite or if we have explicitly setup
-					 * independent streams.
-					 */
-					ic->rex = tick_add_ifset(now_ms, ic->rto);
-				}
-			}
+			ic->rex = tick_add_ifset(now_ms, ic->rto);
 		}
 	}
 }
 
-/* updates both stream ints of a same stream at once */
+/* perform a synchronous send() for the stream interface. The CF_WRITE_NULL and
+ * CF_WRITE_PARTIAL flags are cleared prior to the attempt, and will possibly
+ * be updated in case of success.
+ */
+void si_sync_send(struct stream_interface *si)
+{
+	struct channel *oc = si_oc(si);
+	struct conn_stream *cs;
+
+	oc->flags &= ~(CF_WRITE_NULL|CF_WRITE_PARTIAL);
+
+	if (oc->flags & CF_SHUTW)
+		return;
+
+	if (channel_is_empty(oc))
+		return;
+
+	if (!si_state_in(si->state, SI_SB_CON|SI_SB_RDY|SI_SB_EST))
+		return;
+
+	cs = objt_cs(si->end);
+	if (!cs)
+		return;
+
+	if (cs->flags & (CS_FL_ERROR|CS_FL_ERR_PENDING))
+		return;
+
+	if (cs->conn->flags & CO_FL_ERROR)
+		return;
+
+	if (si_cs_send(cs))
+		si_rx_room_rdy(si_opposite(si));
+}
+
 /* Updates at once the channel flags, and timers of both stream interfaces of a
  * same stream, to complete the work after the analysers, then updates the data
  * layer below. This will ensure that any synchronous update performed at the
  * data layer will be reflected in the channel flags and/or stream-interface.
+ * Note that this does not change the stream interface's current state, though
+ * it updates the previous state to the current one.
  */
 void si_update_both(struct stream_interface *si_f, struct stream_interface *si_b)
 {
 	struct channel *req = si_ic(si_f);
 	struct channel *res = si_oc(si_f);
-	struct conn_stream *cs;
 
 	req->flags &= ~(CF_READ_NULL|CF_READ_PARTIAL|CF_READ_ATTACHED|CF_WRITE_NULL|CF_WRITE_PARTIAL);
 	res->flags &= ~(CF_READ_NULL|CF_READ_PARTIAL|CF_READ_ATTACHED|CF_WRITE_NULL|CF_WRITE_PARTIAL);
 
-	si_f->flags &= ~(SI_FL_ERR|SI_FL_EXP);
-	si_b->flags &= ~(SI_FL_ERR|SI_FL_EXP);
-
 	si_f->prev_state = si_f->state;
 	si_b->prev_state = si_b->state;
 
-	/* front stream-int */
-	cs = objt_cs(si_f->end);
-	if (cs &&
-	    si_f->state == SI_ST_EST &&
-	    !(res->flags & CF_SHUTW) && /* Write not closed */
-	    !channel_is_empty(res) &&
-	    !(cs->flags & CS_FL_ERROR) &&
-	    !(cs->conn->flags & CO_FL_ERROR)) {
-		if (si_cs_send(cs))
-			si_rx_room_rdy(si_b);
-	}
-
-	/* back stream-int */
-	cs = objt_cs(si_b->end);
-	if (cs &&
-	    (si_b->state == SI_ST_EST || si_b->state == SI_ST_CON) &&
-	    !(req->flags & CF_SHUTW) && /* Write not closed */
-	    !channel_is_empty(req) &&
-	    !(cs->flags & (CS_FL_ERROR|CS_FL_ERR_PENDING)) &&
-	    !(cs->conn->flags & CO_FL_ERROR)) {
-		if (si_cs_send(cs))
-			si_rx_room_rdy(si_f);
-	}
-
 	/* let's recompute both sides states */
-	if (si_f->state == SI_ST_EST)
+	if (si_state_in(si_f->state, SI_SB_RDY|SI_SB_EST))
 		si_update(si_f);
 
-	if (si_b->state == SI_ST_EST)
+	if (si_state_in(si_b->state, SI_SB_RDY|SI_SB_EST))
 		si_update(si_b);
 
 	/* stream ints are processed outside of process_stream() and must be
@@ -900,13 +987,12 @@ static void stream_int_shutr_conn(struct stream_interface *si)
 	struct channel *ic = si_ic(si);
 
 	si_rx_shut_blk(si);
-	ic->flags &= ~CF_SHUTR_NOW;
 	if (ic->flags & CF_SHUTR)
 		return;
 	ic->flags |= CF_SHUTR;
 	ic->rex = TICK_ETERNITY;
 
-	if (si->state != SI_ST_EST && si->state != SI_ST_CON)
+	if (!si_state_in(si->state, SI_SB_CON|SI_SB_RDY|SI_SB_EST))
 		return;
 
 	if (si->flags & SI_FL_KILL_CONN)
@@ -951,6 +1037,7 @@ static void stream_int_shutw_conn(struct stream_interface *si)
 	}
 
 	switch (si->state) {
+	case SI_ST_RDY:
 	case SI_ST_EST:
 		/* we have to shut before closing, otherwise some short messages
 		 * may never leave the system, especially when there are remaining
@@ -1007,7 +1094,6 @@ static void stream_int_shutw_conn(struct stream_interface *si)
 	default:
 		si->flags &= ~SI_FL_NOLINGER;
 		si_rx_shut_blk(si);
-		ic->flags &= ~CF_SHUTR_NOW;
 		ic->flags |= CF_SHUTR;
 		ic->rex = TICK_ETERNITY;
 		si->exp = TICK_ETERNITY;
@@ -1023,7 +1109,8 @@ static void stream_int_shutw_conn(struct stream_interface *si)
 static void stream_int_chk_rcv_conn(struct stream_interface *si)
 {
 	/* (re)start reading */
-	tasklet_wakeup(si->wait_event.task);
+	if (si_state_in(si->state, SI_SB_CON|SI_SB_RDY|SI_SB_EST))
+		tasklet_wakeup(si->wait_event.tasklet);
 }
 
 
@@ -1037,7 +1124,8 @@ static void stream_int_chk_snd_conn(struct stream_interface *si)
 	struct channel *oc = si_oc(si);
 	struct conn_stream *cs = __objt_cs(si->end);
 
-	if (unlikely(si->state > SI_ST_EST || (oc->flags & CF_SHUTW)))
+	if (unlikely(!si_state_in(si->state, SI_SB_CON|SI_SB_RDY|SI_SB_EST) ||
+	    (oc->flags & CF_SHUTW)))
 		return;
 
 	if (unlikely(channel_is_empty(oc)))  /* called with nothing to send ! */
@@ -1052,7 +1140,8 @@ static void stream_int_chk_snd_conn(struct stream_interface *si)
 
 	if (cs->flags & (CS_FL_ERROR|CS_FL_ERR_PENDING) || cs->conn->flags & CO_FL_ERROR) {
 		/* Write error on the file descriptor */
-		si->flags |= SI_FL_ERR;
+		if (si->state >= SI_ST_CON)
+			si->flags |= SI_FL_ERR;
 		goto out_wakeup;
 	}
 
@@ -1067,7 +1156,7 @@ static void stream_int_chk_snd_conn(struct stream_interface *si)
 		 */
 		if (((oc->flags & (CF_SHUTW|CF_AUTO_CLOSE|CF_SHUTW_NOW)) ==
 		     (CF_AUTO_CLOSE|CF_SHUTW_NOW)) &&
-		    (si->state == SI_ST_EST)) {
+		    si_state_in(si->state, SI_SB_RDY|SI_SB_EST)) {
 			si_shutw(si);
 			goto out_wakeup;
 		}
@@ -1112,7 +1201,7 @@ static void stream_int_chk_snd_conn(struct stream_interface *si)
 	if (likely((oc->flags & (CF_WRITE_NULL|CF_WRITE_ERROR|CF_SHUTW)) ||
 	          ((oc->flags & CF_WAKE_WRITE) &&
 	           ((channel_is_empty(oc) && !oc->to_forward) ||
-	            si->state != SI_ST_EST)))) {
+	            !si_state_in(si->state, SI_SB_EST))))) {
 	out_wakeup:
 		if (!(si->flags & SI_FL_DONT_WAKE))
 			task_wakeup(si_task(si), TASK_WOKEN_IO);
@@ -1211,6 +1300,8 @@ int si_cs_recv(struct conn_stream *cs)
 			ic->total += ret;
 			cur_read += ret;
 			ic->flags |= CF_READ_PARTIAL;
+			if (si->state == SI_ST_CON)
+				si->state = SI_ST_RDY;
 		}
 
 		if (cs->flags & CS_FL_EOS)
@@ -1256,10 +1347,7 @@ int si_cs_recv(struct conn_stream *cs)
 		 * CS_FL_RCV_MORE on the CS if more space is needed.
 		 */
 		max = channel_recv_max(ic);
-		ret = cs->conn->mux->rcv_buf(cs, &ic->buf, max,
-		          flags |
-		          (co_data(ic) ? CO_RFL_BUF_WET : 0) |
-		          ((channel_recv_limit(ic) < b_size(&ic->buf)) ? CO_RFL_KEEP_RSV : 0));
+		ret = cs->conn->mux->rcv_buf(cs, &ic->buf, max, flags | (co_data(ic) ? CO_RFL_BUF_WET : 0));
 
 		if (cs->flags & CS_FL_WANT_ROOM)
 			si_rx_room_blk(si);
@@ -1274,6 +1362,27 @@ int si_cs_recv(struct conn_stream *cs)
 			break;
 		}
 
+		if (si->flags & SI_FL_L7_RETRY) {
+			struct htx *htx;
+			struct htx_sl *sl;
+
+			htx = htxbuf(&ic->buf);
+			if (htx) {
+				sl = http_get_stline(htx);
+				if (sl && l7_status_match(si_strm(si)->be,
+				    sl->info.res.status)) {
+					/* If we got a status for which we would
+					 * like to retry the request, empty
+					 * the buffer and pretend there's an
+					 * error on the channel.
+					 */
+					ic->flags |= CF_READ_ERROR;
+					htx_reset(htx);
+					return 1;
+				}
+			}
+			si->flags &= ~SI_FL_L7_RETRY;
+		}
 		cur_read += ret;
 
 		/* if we're allowed to directly forward data, we must update ->o */
@@ -1289,6 +1398,8 @@ int si_cs_recv(struct conn_stream *cs)
 
 		ic->flags |= CF_READ_PARTIAL;
 		ic->total += ret;
+		if (si->state == SI_ST_CON)
+			si->state = SI_ST_RDY;
 
 		if ((ic->flags & CF_READ_DONTWAIT) || --read_poll <= 0) {
 			/* we're stopped by the channel's policy */
@@ -1384,7 +1495,7 @@ int si_cs_recv(struct conn_stream *cs)
 		si_rx_endp_more(si);
 	}
 
-	return (cur_read != 0) || si_rx_blocked(si);
+	return (cur_read != 0) || si_rx_blocked(si) || (cs->flags & CS_FL_EOI);
 
  out_shutdown_r:
 	if (conn->flags & CO_FL_CONNECTED) {
@@ -1409,13 +1520,12 @@ static void stream_int_read0(struct stream_interface *si)
 	struct channel *oc = si_oc(si);
 
 	si_rx_shut_blk(si);
-	ic->flags &= ~CF_SHUTR_NOW;
 	if (ic->flags & CF_SHUTR)
 		return;
 	ic->flags |= CF_SHUTR;
 	ic->rex = TICK_ETERNITY;
 
-	if (si->state != SI_ST_EST && si->state != SI_ST_CON)
+	if (!si_state_in(si->state, SI_SB_CON|SI_SB_RDY|SI_SB_EST))
 		return;
 
 	if (oc->flags & CF_SHUTW)
@@ -1441,7 +1551,16 @@ static void stream_int_read0(struct stream_interface *si)
 
 	si_done_get(si);
 
-	si->state = SI_ST_DIS;
+	/* Don't change the state to SI_ST_DIS yet if we're still
+	 * in SI_ST_CON, otherwise it means sess_establish() hasn't
+	 * been called yet, and so the analysers would not run. However
+	 * it's fine to switch to SI_ST_RDY as we have really validated
+	 * the connection.
+	 */
+	if (si->state == SI_ST_EST)
+		si->state = SI_ST_DIS;
+	else if (si->state == SI_ST_CON)
+		si->state = SI_ST_RDY;
 	si->exp = TICK_ETERNITY;
 	return;
 }
@@ -1475,9 +1594,8 @@ void si_applet_wake_cb(struct stream_interface *si)
 	 * appctx but in the case the task is not in runqueue we may have to
 	 * wakeup the appctx immediately.
 	 */
-	if (!task_in_rq(si_task(si)) &&
-	    ((si_rx_endp_ready(si) && !si_rx_blocked(si)) ||
-	     (si_tx_endp_ready(si) && !si_tx_blocked(si))))
+	if ((si_rx_endp_ready(si) && !si_rx_blocked(si)) ||
+	    (si_tx_endp_ready(si) && !si_tx_blocked(si)))
 		appctx_wakeup(si_appctx(si));
 }
 
@@ -1494,7 +1612,6 @@ static void stream_int_shutr_applet(struct stream_interface *si)
 	struct channel *ic = si_ic(si);
 
 	si_rx_shut_blk(si);
-	ic->flags &= ~CF_SHUTR_NOW;
 	if (ic->flags & CF_SHUTR)
 		return;
 	ic->flags |= CF_SHUTR;
@@ -1502,7 +1619,7 @@ static void stream_int_shutr_applet(struct stream_interface *si)
 
 	/* Note: on shutr, we don't call the applet */
 
-	if (si->state != SI_ST_EST && si->state != SI_ST_CON)
+	if (!si_state_in(si->state, SI_SB_CON|SI_SB_RDY|SI_SB_EST))
 		return;
 
 	if (si_oc(si)->flags & CF_SHUTW) {
@@ -1544,6 +1661,7 @@ static void stream_int_shutw_applet(struct stream_interface *si)
 	appctx_wakeup(si_appctx(si));
 
 	switch (si->state) {
+	case SI_ST_RDY:
 	case SI_ST_EST:
 		/* we have to shut before closing, otherwise some short messages
 		 * may never leave the system, especially when there are remaining
@@ -1566,7 +1684,6 @@ static void stream_int_shutw_applet(struct stream_interface *si)
 	default:
 		si->flags &= ~SI_FL_NOLINGER;
 		si_rx_shut_blk(si);
-		ic->flags &= ~CF_SHUTR_NOW;
 		ic->flags |= CF_SHUTR;
 		ic->rex = TICK_ETERNITY;
 		si->exp = TICK_ETERNITY;
